# -*- coding: utf-8 -*-
"""
Azure Resource Manager (ARM) Redis Operations Execution Module

.. versionadded:: 2.0.0

:maintainer: <devops@eitr.tech>
:maturity: new
:depends:
    * `azure <https://pypi.python.org/pypi/azure>`_ >= 4.0.0
    * `azure-common <https://pypi.python.org/pypi/azure-common>`_ >= 1.1.23
    * `azure-mgmt <https://pypi.python.org/pypi/azure-mgmt>`_ >= 4.0.0
    * `azure-mgmt-compute <https://pypi.python.org/pypi/azure-mgmt-compute>`_ >= 4.6.2
    * `azure-mgmt-network <https://pypi.python.org/pypi/azure-mgmt-network>`_ >= 4.0.0
    * `azure-mgmt-redis <https://pypi.org/project/azure-mgmt-redis>`_ >= 6.0.0
    * `azure-mgmt-resource <https://pypi.python.org/pypi/azure-mgmt-resource>`_ >= 2.2.0
    * `azure-mgmt-storage <https://pypi.python.org/pypi/azure-mgmt-storage>`_ >= 2.0.0
    * `azure-mgmt-web <https://pypi.python.org/pypi/azure-mgmt-web>`_ >= 0.35.0
    * `azure-storage <https://pypi.python.org/pypi/azure-storage>`_ >= 0.36.0
    * `msrestazure <https://pypi.python.org/pypi/msrestazure>`_ >= 0.6.1
:platform: linux

:configuration: This module requires Azure Resource Manager credentials to be passed as keyword arguments
    to every function in order to work properly.

    Required provider parameters:

    if using username and password:
      * ``subscription_id``
      * ``username``
      * ``password``

    if using a service principal:
      * ``subscription_id``
      * ``tenant``
      * ``client_id``
      * ``secret``

    Optional provider parameters:

**cloud_environment**: Used to point the cloud driver to different API endpoints, such as Azure GovCloud.
    Possible values:
      * ``AZURE_PUBLIC_CLOUD`` (default)
      * ``AZURE_CHINA_CLOUD``
      * ``AZURE_US_GOV_CLOUD``
      * ``AZURE_GERMAN_CLOUD``

"""
# Python libs
from __future__ import absolute_import
import logging

# Azure libs
HAS_LIBS = False
try:
    import azure.mgmt.network.models  # pylint: disable=unused-import
    from msrest.exceptions import SerializationError
    from msrestazure.azure_exceptions import CloudError

    HAS_LIBS = True
except ImportError:
    pass

__func_alias__ = {"list_": "list"}

log = logging.getLogger(__name__)


async def check_name_availability(hub, ctx, name, **kwargs):
    """
    .. versionadded:: 2.0.0

    Checks that the redis cache name is valid and is not already in use.

    :param name: The name of the Redis cache to check the availability of

    CLI Example:

    .. code-block:: bash

        azurerm.redis.operations.check_name_availability test_name

    """
    result = False
    redconn = await hub.exec.utils.azurerm.get_client(ctx, "redis", **kwargs)

    try:
        avail = redconn.redis.check_name_availability(
            name=name, type="Microsoft.Cache/redis",
        )

        if avail is None:
            result = True
    except CloudError as exc:
        await hub.exec.utils.azurerm.log_cloud_error("redis", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


async def create(
    hub,
    ctx,
    name,
    resource_group,
    location,
    sku,
    redis_configuration=None,
    enable_non_ssl_port=False,
    tenant_settings=None,
    shard_count=None,
    minimum_tls_version=None,
    subnet_id=None,
    static_ip=None,
    zones=None,
    **kwargs,
):
    """
    .. versionadded:: 2.0.0

    Create or replace (overwrite/recreate, with potential downtime) an existing Redis cache.

    :param name: The name of the Redis cache.

    :param resource_group: The name of the resource group.

    :param location: The geo-location where the resource lives.

    :param sku: A dictionary representing the SKU of the Redis cache to deploy. Required parameters include:
        - ``name``: The type of Redis cache to deploy. Possible values include: 'Basic', 'Standard', and 'Premium'.
        - ``family``: The SKU family to use. Possible values include 'C' for Basic/Standard and 'P' for Premium.
        - ``capacity``: The size of the Redis cache to deploy. Possible values include 0, 1, 2, 3, 4, 5, and 6 for the
                        C (Basic/Standard) family and 1, 2, 3, and 4 for the P (Premium) family.

    :param redis_configuration: A dictionary of string key-value pairs that represent all Redis Settings. Some possible
        keys include: rdb-backup-enabled, rdb-storage-connection-string, rdb-backup-frequency, maxmemory-delta,
        maxmemory-policy, notify-keyspace-events, maxmemory-samples, slowlog-log-slower-than, slowlog-max-len,
        list-max-ziplist-entries, list-max-ziplist-value, hash-max-ziplist-entries, hash-max-ziplist-value,
        set-max-intset-entries, zset-max-ziplist-entries, zset-max-ziplist-value, and more.

    :param enable_non_ssl_port: Specifies whether the non-ssl Redis server port (6379) is enabled. Defaults to False.

    :param tenant_settings: A dictionary of tenant settings.

    :param shard_count: The number of shards to be created on a Premium Cluster Cache.

    :param minimum_tls_version: The specified TLS version (or higher) that clients are required to use. Possible values
        include: '1.0', '1.1', and '1.2'.

    :param subnet_id: The full resource ID of a subnet in a virtual network to deploy the Redis cache in. Example
        format: /subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/Microsoft.{Network|ClassicNetwork}/VirtualNetworks/vnet1/subnets/subnet1

    :param static_ip: Static IP address. Required when deploying a Redis cache inside an existing Azure Virtual Network.

    :param zones: A list of availability zones denoting where the resource needs to come from.

    CLI Example:

    .. code-block:: bash

        azurerm.redis.operations.create test_name test_rg test_location test_sku

    """
    redconn = await hub.exec.utils.azurerm.get_client(ctx, "redis", **kwargs)

    try:
        paramsmodel = await hub.exec.utils.azurerm.create_object_model(
            "redis",
            "RedisCreateParameters",
            sku=sku,
            location=location,
            redis_configuration=redis_configuration,
            enable_non_ssl_port=enable_non_ssl_port,
            tenant_settings=tenant_settings,
            shard_count=shard_count,
            minimum_tls_version=minimum_tls_version,
            subnet_id=subnet_id,
            static_ip=static_ip,
            zones=zones,
            **kwargs,
        )
    except TypeError as exc:
        result = {
            "error": "The object model could not be built. ({0})".format(str(exc))
        }
        return result

    try:
        cache = redconn.redis.create(
            name=name, resource_group_name=resource_group, parameters=paramsmodel
        )

        result = cache.result().as_dict()
    except CloudError as exc:
        await hub.exec.utils.azurerm.log_cloud_error("redis", str(exc), **kwargs)
        result = {"error": str(exc)}
    except SerializationError as exc:
        result = {
            "error": "The object model could not be parsed. ({0})".format(str(exc))
        }

    return result


async def delete(hub, ctx, name, resource_group, **kwargs):
    """
    .. versionadded:: 2.0.0

    Deletes a Redis cache.

    :param name: The name of the Redis cache.

    :param resource_group: The name of the resource group.

    CLI Example:

    .. code-block:: bash

        azurerm.redis.operations.delete test_name test_rg

    """
    result = False
    redconn = await hub.exec.utils.azurerm.get_client(ctx, "redis", **kwargs)

    try:
        cache = redconn.redis.delete(name=name, resource_group_name=resource_group)

        result = True
    except CloudError as exc:
        await hub.exec.utils.azurerm.log_cloud_error("redis", str(exc), **kwargs)

    return result


async def export_data(
    hub, ctx, name, resource_group, prefix, container, format=None, **kwargs
):
    """
    .. versionadded:: 2.0.0

    Export data from the redis cache to blobs in a container.

    :param name: The name of the Redis cache.

    :param resource_group: The name of the resource group.

    :param prefix: The prefix to use for exported files.

    :param container: The name of the container to export to.

    :param format: An optional file format.

    CLI Example:

    .. code-block:: bash

        azurerm.redis.operations.export_data test_name test_rg test_prefix test_container

    """
    result = {}
    redconn = await hub.exec.utils.azurerm.get_client(ctx, "redis", **kwargs)

    # Create a ExportRDBParameters object
    try:
        paramsmodel = await hub.exec.utils.azurerm.create_object_model(
            "redis",
            "ExportRDBParameters",
            prefix=prefix,
            container=container,
            format=format,
        )
    except TypeError as exc:
        result = {
            "error": "The object model could not be built. ({0})".format(str(exc))
        }
        return result

    try:
        cache = redconn.redis.export_data(
            name=name, resource_group_name=resource_group, parameters=paramsmodel
        )

        result = cache.result().as_dict()
    except CloudError as exc:
        await hub.exec.utils.azurerm.log_cloud_error("redis", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


async def force_reboot(
    hub, ctx, name, resource_group, reboot_type, shard_id=None, **kwargs
):
    """
    .. versionadded:: 2.0.0

    Reboot specified Redis node(s). This operation requires write permission to the cache resource.
        There can be potential data loss.

    :param name: The name of the redis cache.

    :param resource_group: The name of the resource group.

    :param reboot_type: Which Redis node(s) to reboot. Depending on this value data loss is possible. Possible
        values include: 'PrimaryNode', 'SecondaryNode', 'AllNodes'.

    :param shard_id: If clustering is enabled, the ID of the shard to be rebooted.

    CLI Example:

    .. code-block:: bash

        azurerm.redis.operations.force_reboot test_name test_rg test_type test_id

    """
    result = {}
    redconn = await hub.exec.utils.azurerm.get_client(ctx, "redis", **kwargs)

    try:
        cache = redconn.redis.force_reboot(
            name=name,
            resource_group_name=resource_group,
            reboot_type=reboot_type,
            shard_id=shard_id,
        )

        result = cache.as_dict()
    except CloudError as exc:
        await hub.exec.utils.azurerm.log_cloud_error("redis", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


async def get(hub, ctx, name, resource_group, **kwargs):
    """
    .. versionadded:: 2.0.0

    Gets a Redis cache (resource description).

    :param name: The name of the Redis cache.

    :param resource_group: The name of the resource group.

    CLI Example:

    .. code-block:: bash

        azurerm.redis.operations.get test_name test_rg

    """
    result = {}
    redconn = await hub.exec.utils.azurerm.get_client(ctx, "redis", **kwargs)

    try:
        cache = redconn.redis.get(name=name, resource_group_name=resource_group)

        result = cache.as_dict()
    except CloudError as exc:
        await hub.exec.utils.azurerm.log_cloud_error("redis", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


async def import_data(hub, ctx, name, resource_group, files, format=None, **kwargs):
    """
    .. versionadded:: 2.0.0

    Import data into Redis cache.

    :param name: The name of the Redis cache.

    :param resource_group: The name of the resource group.

    :param files: A list of strings that represent the names of files to import.

    :param format: An optional file format.

    CLI Example:

    .. code-block:: bash

        azurerm.redis.operations.import_data test_name test_rg test_files

    """
    result = {}
    redconn = await hub.exec.utils.azurerm.get_client(ctx, "redis", **kwargs)

    try:
        cache = redconn.redis.import_data(
            name=name, resource_group_name=resource_group, files=files, format=format
        )

        result = cache.result().as_dict()
    except CloudError as exc:
        await hub.exec.utils.azurerm.log_cloud_error("redis", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


async def list_(hub, ctx, **kwargs):
    """
    .. versionadded:: 2.0.0

    Gets all Redis caches in the specified subscription.

    CLI Example:

    .. code-block:: bash

        azurerm.redis.operations.list

    """
    result = {}
    redconn = await hub.exec.utils.azurerm.get_client(ctx, "redis", **kwargs)

    try:
        caches = await hub.exec.utils.azurerm.paged_object_to_list(redconn.redis.list())

        for cache in caches:
            result[cache["name"]] = cache
    except CloudError as exc:
        await hub.exec.utils.azurerm.log_cloud_error("redis", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


async def list_by_resource_group(hub, ctx, resource_group, **kwargs):
    """
    .. versionadded:: 2.0.0

    Lists all Redis caches in a resource group.

    :param resource_group: The name of the resource group.

    CLI Example:

    .. code-block:: bash

        azurerm.redis.operations.list_by_resource_group test_rg

    """
    result = {}
    redconn = await hub.exec.utils.azurerm.get_client(ctx, "redis", **kwargs)

    try:
        caches = await hub.exec.utils.azurerm.paged_object_to_list(
            redconn.redis.list_by_resource_group(resource_group_name=resource_group,)
        )

        for cache in caches:
            result[cache["name"]] = cache
    except CloudError as exc:
        await hub.exec.utils.azurerm.log_cloud_error("redis", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


async def list_keys(hub, ctx, name, resource_group, **kwargs):
    """
    .. versionadded:: 2.0.0

    Retrieve a Redis cache's access keys. This operation requires write permission to the cache resource.

    :param name: The name of the Redis cache.

    :param resource_group: The name of the resource group.

    CLI Example:

    .. code-block:: bash

        azurerm.redis.operations.list_keys test_name test_rg

    """
    result = {}
    redconn = await hub.exec.utils.azurerm.get_client(ctx, "redis", **kwargs)

    try:
        keys = redconn.redis.list_keys(name=name, resource_group_name=resource_group)

        result = keys.as_dict()
    except CloudError as exc:
        await hub.exec.utils.azurerm.log_cloud_error("redis", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


async def list_upgrade_notifications(hub, ctx, name, resource_group, history, **kwargs):
    """
    .. versionadded:: 2.0.0

    Gets any upgrade notifications for a Redis cache.

    :param name: The name of the Redis cache.

    :param resource_group: The name of the resource group.

    :param history: A float representing how many minutes in past to look for upgrade notifications.

    CLI Example:

    .. code-block:: bash

        azurerm.redis.operations.list_upgrade_notifications test_name test_rg test_history

    """
    result = {}
    redconn = await hub.exec.utils.azurerm.get_client(ctx, "redis", **kwargs)

    try:
        notifications = redconn.redis.list_upgrade_notifications(
            name=name, resource_group_name=resource_group, history=history
        )

        result = notifications.as_dict()
    except CloudError as exc:
        await hub.exec.utils.azurerm.log_cloud_error("redis", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


async def regenerate_key(hub, ctx, name, resource_group, key_type, **kwargs):
    """
    .. versionadded:: 2.0.0

    Regenerate Redis cache's access keys. This operation requires write permission to the cache resource.

    :param name: The name of the Redis cache.

    :param resource_group: The name of the resource group.

    :param key_type: The Redis access key to regenerate. Possible values include: 'Primary' and 'Secondary'.

    CLI Example:

    .. code-block:: bash

        azurerm.redis.operations.renegerate_key test_name test_rg test_type

    """
    redconn = await hub.exec.utils.azurerm.get_client(ctx, "redis", **kwargs)

    try:
        keys = redconn.redis.regenerate_key(
            resource_group_name=resource_group, name=name, key_type=key_type, **kwargs
        )

        result = keys.as_dict()
    except CloudError as exc:
        await hub.exec.utils.azurerm.log_cloud_error("redis", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


async def update(
    hub,
    ctx,
    name,
    resource_group,
    sku=None,
    redis_configuration=None,
    enable_non_ssl_port=False,
    tenant_settings=None,
    shard_count=None,
    minimum_tls_version=None,
    **kwargs,
):
    """
    .. versionadded:: 2.0.0

    Update an existing Redis cache.

    :param name: The name of the Redis cache.

    :param resource_group: The name of the resource group.

    :param sku: A dictionary representing the SKU of the Redis cache to deploy. Required parameters include:
        - ``name``: The type of Redis cache to deploy. Possible values include: 'Basic', 'Standard', and 'Premium'.
        - ``family``: The SKU family to use. Possible values include 'C' for Basic/Standard and 'P' for Premium.
        - ``capacity``: The size of the Redis cache to deploy. Possible values include 0, 1, 2, 3, 4, 5, and 6 for the
                        C (Basic/Standard) family and 1, 2, 3, and 4 for the P (Premium) family.

    :param redis_configuration: A dictionary of string key-value pairs that represent all Redis Settings. Some possible
        keys include: rdb-backup-enabled, rdb-storage-connection-string, rdb-backup-frequency, maxmemory-delta,
        maxmemory-policy, notify-keyspace-events, maxmemory-samples, slowlog-log-slower-than, slowlog-max-len,
        list-max-ziplist-entries, list-max-ziplist-value, hash-max-ziplist-entries, hash-max-ziplist-value,
        set-max-intset-entries, zset-max-ziplist-entries, zset-max-ziplist-value, and more.

    :param enable_non_ssl_port: Specifies whether the non-ssl Redis server port (6379) is enabled. Defaults to False.

    :param tenant_settings: A dictionary of tenant settings.

    :param shard_count: The number of shards to be created on a Premium Cluster Cache.

    :param minimum_tls_version: The specified TLS version (or higher) that clients are required to use. Possible values
        include: '1.0', '1.1', and '1.2'.

    CLI Example:

    .. code-block:: bash

        azurerm.redis.operations.update test_name test_rg test_location test_sku

    """
    redconn = await hub.exec.utils.azurerm.get_client(ctx, "redis", **kwargs)

    try:
        paramsmodel = await hub.exec.utils.azurerm.create_object_model(
            "redis",
            "RedisUpdateParameters",
            sku=sku,
            redis_configuration=redis_configuration,
            enable_non_ssl_port=enable_non_ssl_port,
            tenant_settings=tenant_settings,
            shard_count=shard_count,
            minimum_tls_version=minimum_tls_version,
            **kwargs,
        )
    except TypeError as exc:
        result = {
            "error": "The object model could not be built. ({0})".format(str(exc))
        }
        return result

    try:
        cache = redconn.redis.update(
            name=name, resource_group_name=resource_group, parameters=paramsmodel
        )

        result = cache.as_dict()
    except CloudError as exc:
        await hub.exec.utils.azurerm.log_cloud_error("redis", str(exc), **kwargs)
        result = {"error": str(exc)}
    except SerializationError as exc:
        result = {
            "error": "The object model could not be parsed. ({0})".format(str(exc))
        }

    return result



# -*- coding: utf-8 -*-
"""
Azure Resource Manager (ARM) Compute Virtual Machine State Module

.. versionadded:: 1.0.0

.. versionchanged:: 2.0.0

:maintainer: <devops@eitr.tech>
:maturity: new
:depends:
    * `azure <https://pypi.python.org/pypi/azure>`_ >= 4.0.0
    * `azure-common <https://pypi.python.org/pypi/azure-common>`_ >= 1.1.23
    * `azure-mgmt <https://pypi.python.org/pypi/azure-mgmt>`_ >= 4.0.0
    * `azure-mgmt-compute <https://pypi.python.org/pypi/azure-mgmt-compute>`_ >= 4.6.2
    * `azure-mgmt-network <https://pypi.python.org/pypi/azure-mgmt-network>`_ >= 4.0.0
    * `azure-mgmt-resource <https://pypi.python.org/pypi/azure-mgmt-resource>`_ >= 2.2.0
    * `azure-mgmt-storage <https://pypi.python.org/pypi/azure-mgmt-storage>`_ >= 2.0.0
    * `azure-mgmt-web <https://pypi.python.org/pypi/azure-mgmt-web>`_ >= 0.35.0
    * `azure-storage <https://pypi.python.org/pypi/azure-storage>`_ >= 0.36.0
    * `msrestazure <https://pypi.python.org/pypi/msrestazure>`_ >= 0.6.1
:platform: linux

:configuration: This module requires Azure Resource Manager credentials to be passed as a dictionary of
    keyword arguments to the ``connection_auth`` parameter in order to work properly. Since the authentication
    parameters are sensitive, it's recommended to pass them to the states via pillar.

    Required provider parameters:

    if using username and password:
      * ``subscription_id``
      * ``username``
      * ``password``

    if using a service principal:
      * ``subscription_id``
      * ``tenant``
      * ``client_id``
      * ``secret``

    Optional provider parameters:

    **cloud_environment**: Used to point the cloud driver to different API endpoints, such as Azure GovCloud. Possible values:
      * ``AZURE_PUBLIC_CLOUD`` (default)
      * ``AZURE_CHINA_CLOUD``
      * ``AZURE_US_GOV_CLOUD``
      * ``AZURE_GERMAN_CLOUD``

    Example Pillar for Azure Resource Manager authentication:

    .. code-block:: yaml

        azurerm:
            user_pass_auth:
                subscription_id: 3287abc8-f98a-c678-3bde-326766fd3617
                username: fletch
                password: 123pass
            mysubscription:
                subscription_id: 3287abc8-f98a-c678-3bde-326766fd3617
                tenant: ABCDEFAB-1234-ABCD-1234-ABCDEFABCDEF
                client_id: ABCDEFAB-1234-ABCD-1234-ABCDEFABCDEF
                secret: XXXXXXXXXXXXXXXXXXXXXXXX
                cloud_environment: AZURE_PUBLIC_CLOUD

    Example states using Azure Resource Manager authentication:

    .. code-block:: jinja

        Ensure virtual machine exists:
          azurerm.compute.virtual_machine.present:
            - name: idem-vm01
            - resource_group: idem
            - vm_size: Standard_B1s
            - virtual_network: vnet1
            - subnet: default
            - allocate_public_ip: True
            - ssh_public_keys:
                - /home/myuser/.ssh/id_rsa.pub
            - tags:
                contact_name: Elmer Fudd Gantry
            - connection_auth: {{ profile }}

        Ensure virtual machine is absent:
            azurerm.compute.virtual_machine.absent:
                - name: idem-vm01
                - resource_group: idem
                - connection_auth: {{ profile }}

"""
# Python libs
from __future__ import absolute_import
import logging
import operator

# Azure libs
HAS_LIBS = False
try:
    from msrestazure.tools import parse_resource_id

    HAS_LIBS = True
except ImportError:
    pass


log = logging.getLogger(__name__)

TREQ = {
    "present": {
        "require": [
            "states.azurerm.resource.group.present",
            "states.azurerm.network.virtual_network.present",
            "states.azurerm.network.virtual_network.subnet_present",
            "states.azurerm.network.network_interface.present",
            "states.azurerm.compute.availability_set.present",
        ]
    },
}


async def present(
    hub,
    ctx,
    name,
    resource_group,
    vm_size,
    admin_username="idem",
    os_disk_create_option="FromImage",
    os_disk_size_gb=30,
    ssh_public_keys=None,
    disable_password_auth=None,
    custom_data=None,
    allow_extensions=None,
    enable_automatic_updates=None,
    time_zone=None,
    allocate_public_ip=False,
    create_interfaces=True,
    network_resource_group=None,
    virtual_network=None,
    subnet=None,
    network_interfaces=None,
    os_managed_disk=None,
    os_disk_vhd_uri=None,
    os_disk_image_uri=None,
    os_type=None,
    os_disk_name=None,
    os_disk_caching=None,
    os_write_accel=None,
    os_ephemeral_disk=None,
    ultra_ssd_enabled=None,
    image=None,
    boot_diags_enabled=None,
    diag_storage_uri=None,
    admin_password=None,
    force_admin_password=False,
    max_price=None,
    provision_vm_agent=True,
    userdata_file=None,
    userdata=None,
    enable_disk_enc=False,
    disk_enc_keyvault=None,
    disk_enc_volume_type=None,
    disk_enc_kek_url=None,
    data_disks=None,
    tags=None,
    connection_auth=None,
    **kwargs,
):
    """
    .. versionadded:: 1.0.0

    .. versionchanged:: 2.0.0

    Ensure a virtual machine exists.

    :param name: The virtual machine to ensure is present

    :param resource_group: The resource group name assigned to the virtual machine.

    :param vm_size: The size of the virtual machine.

    :param admin_username: Specifies the name of the administrator account.

    :param os_disk_create_option: (attach, from_image, or empty) Specifies how the virtual machine should be created.
        The "attach" value is used when you are using a specialized disk to create the virtual machine. The "from_image"
        value is used when you are using an image to create the virtual machine. If you are using a platform image, you
        also use the image_reference element. If you are using a marketplace image, you also use the plan element.

    :param os_disk_size_gb: Specifies the size of an empty OS disk in gigabytes. This element can be used to overwrite
        the size of the disk in a virtual machine image.

    :param ssh_public_keys: The list of SSH public keys used to authenticate with Linux based VMs.

    :param disable_password_auth: (only on Linux) Specifies whether password authentication should be disabled when SSH
        public keys are provided.

    :param custom_data: (only on Linux) Specifies a base-64 encoded string of custom data for cloud-init (not user-data
        scripts). The base-64 encoded string is decoded to a binary array that is saved as a file on the Virtual
        Machine. The maximum length of the binary array is 65535 bytes. For using cloud-init for your VM, see `Using
        cloud-init to customize a Linux VM during creation
        <https://docs.microsoft.com/en-us/azure/virtual-machines/linux/using-cloud-init>`_

    :param allow_extensions: Specifies whether extension operations should be allowed on the virtual machine. This may
        only be set to False when no extensions are present on the virtual machine.

    :param enable_automatic_updates: (only on Windows) Indicates whether automatic updates are enabled for the Windows
        virtual machine. Default value is true. For virtual machine scale sets, this property can be updated and updates
        will take effect on OS reprovisioning.

    :param time_zone: (only on Windows) Specifies the time zone of the virtual machine. e.g. "Pacific Standard Time"

    :param allocate_public_ip: Create and attach a public IP object to the VM.

    :param create_interfaces: Create network interfaces to attach to the VM if none are provided.

    :param network_resource_group: Specify the resource group of the network components referenced in this module.

    :param virtual_network: Virtual network for the subnet which will contain the network interfaces.

    :param subnet: Subnet to which the network interfaces will be attached.

    :param network_interfaces: A list of network interface references ({"id": "/full/path/to/object"}) to attach.

    :param os_managed_disk: A managed disk resource ID or dictionary containing the managed disk parameters. If a
        dictionary is provided, "storage_account_type" can be passed in additional to the "id". Storage account type for
        the managed disk can include: 'Standard_LRS', 'Premium_LRS', 'StandardSSD_LRS', 'UltraSSD_LRS'. NOTE:
        UltraSSD_LRS can only be used with data disks.

    :param os_disk_vhd_uri: The virtual hard disk for the OS ({"uri": "/full/path/to/object"}).

    :param os_disk_image_uri: The source user image virtual hard disk ({"uri": "/full/path/to/object"}). The virtual
        hard disk will be copied before being attached to the virtual machine. If SourceImage is provided, the
        destination virtual hard drive must not exist.

    :param os_type: (linux or windows) This property allows you to specify the type of the OS that is included in the
        disk if creating a VM from user-image or a specialized VHD.

    :param os_disk_name: The OS disk name.

    :param os_disk_caching: (read_only, read_write, or none) Specifies the caching requirements. Defaults
        to "None" for Standard storage and "ReadOnly" for Premium storage.

    :param os_write_accel: Boolean value specifies whether write accelerator should be enabled or disabled on the disk.

    :param os_ephemeral_disk: Boolean value to enable ephemeral "diff" OS disk. `Ephemeral OS disks
        <https://docs.microsoft.com/en-us/azure/virtual-machines/linux/ephemeral-os-disks>`_ are created on the local
        virtual machine (VM) storage and not saved to the remote Azure Storage.

    :param ultra_ssd_enabled: The flag that enables or disables a capability to have one or more managed data disks with
        UltraSSD_LRS storage account type on the VM or VMSS. Managed disks with storage account type UltraSSD_LRS can be
        added to a virtual machine or virtual machine scale set only if this property is enabled.

    :param image: A pipe-delimited representation of an image to use, in the format of "publisher|offer|sku|version".
        Examples - "OpenLogic|CentOS|7.7|latest" or "Canonical|UbuntuServer|18.04-LTS|latest"

    :param boot_diags_enabled: Enables boots diagnostics on the Virtual Machine. Required for use of the
        diag_storage_uri parameter.

    :param diag_storage_uri: Enables boots diagnostics on the Virtual Machine by passing the URI of the storage account
        to use for placing the console output and screenshot.

    :param admin_password: Specifies the password of the administrator account. Note that there are minimum length,
        maximum length, and complexity requirements imposed on this password. See the Azure documentation for details.

    :param force_admin_password: A Boolean flag that represents whether or not the admin password should be updated.
        If it is set to True, then the admin password will be updated if the virtual machine already exists. If it is
        set to False, then the password will not be updated unless other parameters also need to be updated.
        Defaults to False.

    :param provision_vm_agent: Indicates whether virtual machine agent should be provisioned on the virtual machine.
        When this property is not specified in the request body, default behavior is to set it to true. This will ensure
        that VM Agent is installed on the VM so that extensions can be added to the VM later. If attempting to set this
        value, os_type should also be set in order to ensure the proper OS configuration is used.

    :param userdata_file: This parameter can contain a local or web path for a userdata script. If a local file is used,
        then the contents of that file will override the contents of the userdata parameter. If a web source is used,
        then the userdata parameter should contain the command to execute the script file. For instance, if a file
        location of https://raw.githubusercontent.com/saltstack/salt-bootstrap/stable/bootstrap-salt.sh is used then the
        userdata parameter would contain "./bootstrap-salt.sh" along with any desired arguments. Note that PowerShell
        execution policy may cause issues here. For PowerShell files, considered signed scripts or the more insecure
        "powershell -ExecutionPolicy Unrestricted -File ./bootstrap-salt.ps1" addition to the command.

    :param userdata: This parameter is used to pass text to be executed on a system. The native shell will be used on a
        given host operating system.

    :param max_price: Specifies the maximum price you are willing to pay for a Azure Spot VM/VMSS. This price is in US
        Dollars. This price will be compared with the current Azure Spot price for the VM size. Also, the prices are
        compared at the time of create/update of Azure Spot VM/VMSS and the operation will only succeed if max_price is
        greater than the current Azure Spot price. The max_price will also be used for evicting a Azure Spot VM/VMSS if
        the current Azure Spot price goes beyond the maxPrice after creation of VM/VMSS. Possible values are any decimal
        value greater than zero (example: 0.01538) or -1 indicates default price to be up-to on-demand. You can set the
        max_price to -1 to indicate that the Azure Spot VM/VMSS should not be evicted for price reasons. Also, the
        default max price is -1 if it is not provided by you.

    :param availability_set: Specifies information about the availability set that the virtual machine should be
        assigned to. Virtual machines specified in the same availability set are allocated to different nodes to
        maximize availability. For more information about availability sets, see `Manage the availability of virtual
        machines <https://docs.microsoft.com/azure/virtual-machines/virtual-machines-windows-manage-availability>`_.
        Currently, a VM can only be added to availability set at creation time. An existing VM cannot be added to an
        availability set. (resource ID path)

    :param virtual_machine_scale_set: Specifies information about the virtual machine scale set that the virtual machine
        should be assigned to. Virtual machines specified in the same virtual machine scale set are allocated to
        different nodes to maximize availability. Currently, a VM can only be added to virtual machine scale set at
        creation time. An existing VM cannot be added to a virtual machine scale set. This property cannot exist along
        with a non-null availability_set reference. (resource ID path)

    :param proximity_placement_group: Specifies information about the proximity placement group that the virtual machine
        should be assigned to.

    :param priority: (low or regular) Specifies the priority for the virtual machine.

    :param eviction_policy: (deallocate or delete) Specifies the eviction policy for the Azure Spot virtual machine.

    :param license_type: (Windows_Client or Windows_Server) Specifies that the image or disk that is being used was
        licensed on-premises. This element is only used for images that contain the Windows Server operating system.

    :param zones: A list of the virtual machine zones.

    :param tags: A dictionary of strings can be passed as tag metadata to the virtual machine object.

    :param connection_auth: A dictionary with subscription and authentication parameters to be used in connecting to
        the Azure Resource Manager API.

    Virtual Machine Disk Encryption:
        If you would like to enable disk encryption within the virtual machine you must set the enable_disk_enc
        parameter to True. Disk encryption utilizes a VM published by Microsoft.Azure.Security of extension type
        AzureDiskEncryptionForLinux or AzureDiskEncryption, depending on your virtual machine OS. More information
        about Disk Encryption and its requirements can be found in the links below.

        Disk Encryption for Windows Virtual Machines:
        https://docs.microsoft.com/en-us/azure/virtual-machines/windows/disk-encryption-overview

        Disk Encryption for Linux Virtual Machines:
        https://docs.microsoft.com/en-us/azure/virtual-machines/linux/disk-encryption-overview

        The following parameters may be used to implement virtual machine disk encryption:
        :param enable_disk_enc: This boolean flag will represent whether disk encryption has been enabled for the
            virtual machine. This is a required parameter.

        :param disk_enc_keyvault: The resource ID of the key vault containing the disk encryption key, which is a
            Key Vault Secret. This is a required parameter.

        :param disk_enc_volume_type: The volume type(s) that will be encrypted. Possible values include: 'OS',
            'Data', and 'All'. This is a required parameter.

        :param disk_enc_kek_url: The Key Identifier URL for a Key Encryption Key (KEK). The KEK is used as an
            additional layer of security for encryption keys. Azure Disk Encryption will use the KEK to wrap the
            encryption secrets before writing to the Key Vault. The KEK must be in the same vault as the encryption
            secrets. This is an optional parameter.

    Attaching Data Disks:
        Data disks can be attached by passing a list of dictionaries in the data_disks parameter. The dictionaries in
        the list can have the following parameters.

        :param lun: (optional int) Specifies the logical unit number of the data disk. This value is used to identify
            data disks within the VM and therefore must be unique for each data disk attached to a VM. If not
            provided, we increment the lun designator based upon the index within the provided list of disks.

        :param name: (optional str) The disk name. Defaults to "{vm_name}-datadisk{lun}"

        :param vhd: (optional str or dict) Virtual hard disk to use. If a URI string is provided, it will be nested
            under a "uri" key in a dictionary as expected by the SDK.

        :param image: (optional str or dict) The source user image virtual hard disk. The virtual hard disk will be
            copied before being attached to the virtual machine. If image is provided, the destination virtual hard
            drive must not exist. If a URI string is provided, it will be nested under a "uri" key in a dictionary as
            expected by the SDK.

        :param caching: (optional str - read_only, read_write, or none) Specifies the caching requirements. Defaults to
            "None" for Standard storage and "ReadOnly" for Premium storage.

        :param write_accelerator_enabled: (optional bool - True or False) Specifies whether write accelerator should be
            enabled or disabled on the disk.

        :param create_option: (optional str - attach, from_image, or empty) Specifies how the virtual machine should be
            created. The "attach" value is used when you are using a specialized disk to create the virtual machine. The
            "from_image" value is used when you are using an image to create the virtual machine. If you are using a
            platform image, you also use the image_reference element. If you are using a marketplace image, you also use
            the plan element.

        :param disk_size_gb: (optional int) Specifies the size of an empty data disk in gigabytes. This element can be
            used to overwrite the size of the disk in a virtual machine image.

        :param managed_disk: (optional str or dict) The managed disk parameters. If an ID string is provided, it will
            be nested under an "id" key in a dictionary as expected by the SDK. If a dictionary is provided, the
            "storage_account_type" parameter can be passed (accepts (Standard|Premium)_LRS or (Standard|Ultra)SSD_LRS).

    Example usage:

    .. code-block:: yaml

        Ensure virtual machine exists:
            azurerm.compute.virtual_machine.present:
                - name: idem-vm01
                - resource_group: idem
                - vm_size: Standard_B1s
                - virtual_network: vnet1
                - subnet: default
                - allocate_public_ip: True
                - ssh_public_keys:
                    - /home/myuser/.ssh/id_rsa.pub
                - tags:
                    contact_name: Elmer Fudd Gantry
                - connection_auth: {{ profile }}

    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        if ctx["acct"]:
            connection_auth = ctx["acct"]
        else:
            ret[
                "comment"
            ] = "Connection information must be specified via acct or connection_auth dictionary!"
            return ret

    vm = await hub.exec.azurerm.compute.virtual_machine.get(
        ctx, name, resource_group, azurerm_log_level="info", **connection_auth
    )

    new_vm = True

    if "error" not in vm:
        new_vm = False

        tag_changes = await hub.exec.utils.dictdiffer.deep_diff(
            vm.get("tags", {}), tags or {}
        )
        if tag_changes:
            ret["changes"]["tags"] = tag_changes

        if vm_size.lower() != vm["hardware_profile"]["vm_size"].lower():
            ret["changes"]["vm_size"] = {
                "old": vm["hardware_profile"]["vm_size"].lower(),
                "new": vm_size.lower(),
            }

        if boot_diags_enabled is not None:
            if boot_diags_enabled != vm.get("diagnostics_profile", {}).get(
                "boot_diagnostics", {}
            ).get("enabled", False):
                ret["changes"]["boot_diags_enabled"] = {
                    "old": vm.get("diagnostics_profile", {})
                    .get("boot_diagnostics", {})
                    .get("enabled", False),
                    "new": boot_diags_enabled,
                }

        if diag_storage_uri:
            if diag_storage_uri != vm.get("diagnostics_profile", {}).get(
                "boot_diagnostics", {}
            ).get("storage_uri"):
                ret["changes"]["diag_storage_uri"] = {
                    "old": vm.get("diagnostics_profile", {})
                    .get("boot_diagnostics", {})
                    .get("storage_uri"),
                    "new": diag_storage_uri,
                }

        if max_price:
            if max_price != vm.get("billing_profile", {}).get("max_price"):
                ret["changes"]["max_price"] = {
                    "old": vm.get("billing_profile", {}).get("max_price"),
                    "new": max_price,
                }

        if allow_extensions is not None:
            if allow_extensions != vm.get("os_profile", {}).get(
                "allow_extension_operations", True
            ):
                ret["changes"]["allow_extensions"] = {
                    "old": vm.get("os_profile", {}).get(
                        "allow_extension_operations", True
                    ),
                    "new": allow_extensions,
                }

        if os_write_accel is not None:
            if os_write_accel != vm.get("storage_profile", {}).get("os_disk", {}).get(
                "write_accelerator_enabled"
            ):
                ret["changes"]["os_write_accel"] = {
                    "old": vm.get("storage_profile", {})
                    .get("os_disk", {})
                    .get("write_accelerator_enabled"),
                    "new": os_write_accel,
                }

        if os_disk_caching is not None:
            if os_disk_caching != vm.get("storage_profile", {}).get("os_disk", {}).get(
                "caching"
            ):
                ret["changes"]["os_disk_caching"] = {
                    "old": vm.get("storage_profile", {})
                    .get("os_disk", {})
                    .get("caching"),
                    "new": os_disk_caching,
                }

        if ultra_ssd_enabled is not None:
            if ultra_ssd_enabled != vm.get("additional_capabilities", {}).get(
                "ultra_ssd_enabled"
            ):
                ret["changes"]["ultra_ssd_enabled"] = {
                    "old": vm.get("additional_capabilities", {}).get(
                        "ultra_ssd_enabled"
                    ),
                    "new": ultra_ssd_enabled,
                }

        if provision_vm_agent is not None:
            if vm.get("os_profile", {}).get("linux_configuration", {}):
                if provision_vm_agent != vm["os_profile"]["linux_configuration"].get(
                    "provision_vm_agent", True
                ):
                    ret["changes"]["provision_vm_agent"] = {
                        "old": vm["os_profile"]["linux_configuration"].get(
                            "provision_vm_agent", True
                        ),
                        "new": provision_vm_agent,
                    }
            if vm.get("os_profile", {}).get("windows_configuration", {}):
                if provision_vm_agent != vm["os_profile"]["windows_configuration"].get(
                    "provision_vm_agent", True
                ):
                    ret["changes"]["provision_vm_agent"] = {
                        "old": vm["os_profile"]["windows_configuration"].get(
                            "provision_vm_agent", True
                        ),
                        "new": provision_vm_agent,
                    }

        if time_zone:
            if time_zone != vm.get("os_profile", {}).get(
                "windows_configuration", {}
            ).get("time_zone", True):
                ret["changes"]["time_zone"] = {
                    "old": vm.get("os_profile", {})
                    .get("windows_configuration", {})
                    .get("time_zone", True),
                    "new": time_zone,
                }

        if enable_automatic_updates is not None:
            if enable_automatic_updates != vm.get("os_profile", {}).get(
                "windows_configuration", {}
            ).get("enable_automatic_updates", True):
                ret["changes"]["enable_automatic_updates"] = {
                    "old": vm.get("os_profile", {})
                    .get("windows_configuration", {})
                    .get("enable_automatic_updates", True),
                    "new": enable_automatic_updates,
                }

        if data_disks is not None:
            existing_disks = vm.get("storage_profile", {}).get("data_disks", [])

            if len(existing_disks) != len(data_disks):
                ret["changes"]["data_disks"] = {
                    "old": existing_disks,
                    "new": data_disks,
                }
            else:
                for idx, disk in enumerate(data_disks):
                    for key in disk:
                        if isinstance(disk[key], dict) and isinstance(
                            existing_disks[idx].get(key), dict
                        ):
                            for k in disk[key]:
                                if disk[key][k] != existing_disks[idx][key].get(k):
                                    ret["changes"]["data_disks"] = {
                                        "old": existing_disks,
                                        "new": data_disks,
                                    }
                        else:
                            if disk[key] != existing_disks[idx].get(key):
                                ret["changes"]["data_disks"] = {
                                    "old": existing_disks,
                                    "new": data_disks,
                                }

        if enable_disk_enc:
            extensions = vm.get("resources", [])
            disk_enc_exists = False
            for extension in extensions:
                if (
                    extension.get("virtual_machine_extension_type")
                    == "AzureDiskEncryptionForLinux"
                    or extension.get("virtual_machine_extension_type")
                    == "AzureDiskEncryption"
                ):
                    disk_enc_exists = True
                    break

            if not disk_enc_exists:
                ret["changes"]["enable_disk_enc"] = {"old": False, "new": True}
                if disk_enc_keyvault:
                    ret["changes"]["disk_enc_keyvault"] = {"new": disk_enc_keyvault}
                if disk_enc_volume_type:
                    ret["changes"]["disk_enc_volume_type"] = {
                        "new": disk_enc_volume_type
                    }
                if disk_enc_kek_url:
                    ret["changes"]["disk_enc_kek_url"] = {"new": disk_enc_kek_url}

        if admin_password and force_admin_password:
            ret["changes"]["admin_password"] = {"new": "REDACTED"}
        elif ret["changes"]:
            ret["changes"]["admin_password"] = {"new": "REDACTED"}

        if not ret["changes"]:
            ret["result"] = True
            ret["comment"] = "Virtual machine {0} is already present.".format(name)
            return ret

        if ctx["test"]:
            ret["result"] = None
            ret["comment"] = "Virtual machine {0} would be updated.".format(name)
            return ret
    else:
        ret["changes"] = {
            "old": {},
            "new": {
                "name": name,
                "resource_group": resource_group,
                "vm_size": vm_size,
                "os_disk_create_option": os_disk_create_option,
                "os_disk_size_gb": os_disk_size_gb,
            },
        }

        if ssh_public_keys:
            ret["changes"]["new"]["ssh_public_keys"] = ssh_public_keys
        if disable_password_auth is not None:
            ret["changes"]["new"]["disable_password_auth"] = disable_password_auth
        if custom_data:
            ret["changes"]["new"]["custom_data"] = custom_data
        if allow_extensions is not None:
            ret["changes"]["new"]["allow_extensions"] = allow_extensions
        if enable_automatic_updates is not None:
            ret["changes"]["new"]["enable_automatic_updates"] = enable_automatic_updates
        if time_zone:
            ret["changes"]["new"]["time_zone"] = time_zone
        if allocate_public_ip is not None:
            ret["changes"]["new"]["allocate_public_ip"] = allocate_public_ip
        if create_interfaces is not None:
            ret["changes"]["new"]["create_interfaces"] = create_interfaces
        if network_resource_group:
            ret["changes"]["new"]["network_resource_group"] = network_resource_group
        if virtual_network:
            ret["changes"]["new"]["virtual_network"] = virtual_network
        if subnet:
            ret["changes"]["new"]["subnet"] = subnet
        if network_interfaces:
            ret["changes"]["new"]["network_interfaces"] = network_interfaces
        if os_managed_disk:
            ret["changes"]["new"]["os_managed_disk"] = os_managed_disk
        if os_disk_vhd_uri:
            ret["changes"]["new"]["os_disk_vhd_uri"] = os_disk_vhd_uri
        if os_disk_image_uri:
            ret["changes"]["new"]["os_disk_image_uri"] = os_disk_image_uri
        if os_type:
            ret["changes"]["new"]["os_type"] = os_type
        if os_disk_name:
            ret["changes"]["new"]["os_disk_name"] = os_disk_name
        if os_disk_caching:
            ret["changes"]["new"]["os_disk_caching"] = os_disk_caching
        if os_write_accel is not None:
            ret["changes"]["new"]["os_write_accel"] = os_write_accel
        if os_ephemeral_disk is not None:
            ret["changes"]["new"]["os_ephemeral_disk"] = os_ephemeral_disk
        if ultra_ssd_enabled is not None:
            ret["changes"]["new"]["ultra_ssd_enabled"] = ultra_ssd_enabled
        if image:
            ret["changes"]["new"]["image"] = image
        if boot_diags_enabled is not None:
            ret["changes"]["new"]["boot_diags_enabled"] = boot_diags_enabled
        if diag_storage_uri:
            ret["changes"]["new"]["diag_storage_uri"] = diag_storage_uri
        if admin_password:
            ret["changes"]["new"]["admin_password"] = admin_password
        if max_price:
            ret["changes"]["new"]["max_price"] = max_price
        if provision_vm_agent is not None:
            ret["changes"]["new"]["provision_vm_agent"] = provision_vm_agent
        if userdata_file:
            ret["changes"]["new"]["userdata_file"] = userdata_file
        if userdata:
            ret["changes"]["new"]["userdata"] = userdata
        if enable_disk_enc is not None:
            ret["changes"]["new"]["enable_disk_enc"] = enable_disk_enc
        if disk_enc_keyvault:
            ret["changes"]["new"]["disk_enc_keyvault"] = disk_enc_keyvault
        if disk_enc_volume_type:
            ret["changes"]["new"]["disk_enc_volume_type"] = disk_enc_volume_type
        if disk_enc_kek_url:
            ret["changes"]["new"]["disk_enc_kek_url"] = disk_enc_kek_url
        if data_disks:
            ret["changes"]["new"]["data_disks"] = data_disks
        if tags is not None:
            ret["changes"]["new"]["tags"] = tags

    if ctx["test"]:
        ret["result"] = None
        ret["comment"] = "Virtual machine {0} would be created.".format(name)
        return ret

    vm_kwargs = kwargs.copy()
    vm_kwargs.update(connection_auth)

    vm = await hub.exec.azurerm.compute.virtual_machine.create_or_update(
        ctx=ctx,
        name=name,
        resource_group=resource_group,
        vm_size=vm_size,
        admin_username=admin_username,
        os_disk_create_option=os_disk_create_option,
        os_disk_size_gb=os_disk_size_gb,
        ssh_public_keys=ssh_public_keys,
        disable_password_auth=disable_password_auth,
        custom_data=custom_data,
        allow_extensions=allow_extensions,
        enable_automatic_updates=enable_automatic_updates,
        time_zone=time_zone,
        allocate_public_ip=allocate_public_ip,
        create_interfaces=create_interfaces,
        network_resource_group=network_resource_group,
        virtual_network=virtual_network,
        subnet=subnet,
        network_interfaces=network_interfaces,
        os_managed_disk=os_managed_disk,
        os_disk_vhd_uri=os_disk_vhd_uri,
        os_disk_image_uri=os_disk_image_uri,
        os_type=os_type,
        os_disk_name=os_disk_name,
        os_disk_caching=os_disk_caching,
        os_write_accel=os_write_accel,
        os_ephemeral_disk=os_ephemeral_disk,
        ultra_ssd_enabled=ultra_ssd_enabled,
        image=image,
        boot_diags_enabled=boot_diags_enabled,
        diag_storage_uri=diag_storage_uri,
        admin_password=admin_password,
        max_price=max_price,
        provision_vm_agent=provision_vm_agent,
        userdata_file=userdata_file,
        userdata=userdata,
        enable_disk_enc=enable_disk_enc,
        disk_enc_keyvault=disk_enc_keyvault,
        disk_enc_volume_type=disk_enc_volume_type,
        disk_enc_kek_url=disk_enc_kek_url,
        data_disks=data_disks,
        tags=tags,
        **vm_kwargs,
    )

    if new_vm:
        ret["changes"] = {"old": {}, "new": vm}

    if "error" not in vm:
        ret["result"] = True
        ret["comment"] = "Virtual machine {0} has been created.".format(name)
        return ret

    ret["comment"] = "Failed to create virtual machine {0}! ({1})".format(
        name, vm.get("error")
    )
    if not ret["result"]:
        ret["changes"] = {}
    return ret


async def absent(
    hub,
    ctx,
    name,
    resource_group,
    cleanup_osdisks=False,
    cleanup_datadisks=False,
    cleanup_interfaces=False,
    cleanup_public_ips=False,
    connection_auth=None,
    **kwargs,
):
    """
    .. versionadded:: 1.0.0

    .. versionchanged:: 2.0.0

    Ensure a virtual machine does not exist in a resource group.

    :param name:
        Name of the virtual machine.

    :param resource_group:
        Name of the resource group containing the virtual machine.

    :param cleanup_osdisks:
        Enable deletion of the operating system disk attached to the virtual machine.

    :param cleanup_datadisks:
        Enable deletion of ALL of the data disks attached to the virtual machine.

    :param cleanup_interfaces:
        Enable deletion of ALL of the network interfaces attached to the virtual machine.

    :param cleanup_public_ips:
        Enable deletion of ALL of the public IP addresses directly attached to the virtual machine.

    :param connection_auth:
        A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.

    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        if ctx["acct"]:
            connection_auth = ctx["acct"]
        else:
            ret[
                "comment"
            ] = "Connection information must be specified via acct or connection_auth dictionary!"
            return ret

    vm = await hub.exec.azurerm.compute.virtual_machine.get(
        ctx, name, resource_group, azurerm_log_level="info", **connection_auth
    )

    if "error" in vm:
        ret["result"] = True
        ret["comment"] = "Virtual machine {0} was not found.".format(name)
        return ret

    elif ctx["test"]:
        ret["comment"] = "Virtual machine {0} would be deleted.".format(name)
        ret["result"] = None
        ret["changes"] = {
            "old": vm,
            "new": {},
        }
        return ret

    deleted = await hub.exec.azurerm.compute.virtual_machine.delete(
        ctx, name, resource_group, **connection_auth
    )

    if deleted:
        if cleanup_osdisks:
            vm["cleanup_osdisks"] = True
            os_disk = vm["storage_profile"]["os_disk"]
            if os_disk.get("managed_disk", {}).get("id"):
                disk_link = os_disk["managed_disk"]["id"]
                try:
                    disk_dict = parse_resource_id(disk_link)
                    disk_name = disk_dict["name"]
                    disk_group = disk_dict["resource_group"]
                except KeyError as exc:
                    log.error("This isn't a valid disk resource: %s", os_disk)

                deleted_disk = await hub.exec.azurerm.compute.disk.delete(
                    ctx,
                    disk_name,
                    disk_group,
                    azurerm_log_level="info",
                    **connection_auth,
                )

                if not deleted_disk:
                    log.error("Unable to delete disk: %s", disk_link)

        if cleanup_datadisks:
            vm["cleanup_datadisks"] = True
            for disk in vm["storage_profile"].get("data_disks", []):
                if disk.get("managed_disk", {}).get("id"):
                    disk_link = disk["managed_disk"]["id"]
                    try:
                        disk_dict = parse_resource_id(disk_link)
                        disk_name = disk_dict["name"]
                        disk_group = disk_dict["resource_group"]
                    except KeyError as exc:
                        log.error("This isn't a valid disk resource: %s", os_disk)
                        continue

                    deleted_disk = await hub.exec.azurerm.compute.disk.delete(
                        ctx,
                        disk_name,
                        disk_group,
                        azurerm_log_level="info",
                        **connection_auth,
                    )

                    if not deleted_disk:
                        log.error("Unable to delete disk: %s", disk_link)

        if cleanup_interfaces:
            vm["cleanup_interfaces"] = True
            for nic_link in vm.get("network_profile", {}).get("network_interfaces", []):
                try:
                    nic_dict = parse_resource_id(nic_link["id"])
                    nic_name = nic_dict["name"]
                    nic_group = nic_dict["resource_group"]
                except KeyError as exc:
                    log.error(
                        "This isn't a valid network interface subresource: %s", nic_link
                    )
                    continue

                nic = await hub.exec.azurerm.network.network_interface.get(
                    ctx,
                    nic_name,
                    nic_group,
                    azurerm_log_level="info",
                    **connection_auth,
                )

                deleted_nic = await hub.exec.azurerm.network.network_interface.delete(
                    ctx,
                    nic_name,
                    nic_group,
                    azurerm_log_level="info",
                    **connection_auth,
                )

                if cleanup_public_ips:
                    vm["cleanup_public_ips"] = True
                    for ipc in nic.get("ip_configurations", []):
                        if "public_ip_address" not in ipc:
                            continue

                        try:
                            pip_dict = parse_resource_id(ipc["public_ip_address"]["id"])
                            pip_name = pip_dict["name"]
                            pip_group = pip_dict["resource_group"]
                        except KeyError as exc:
                            log.error(
                                "This isn't a valid public IP subresource: %s",
                                ipc.get("public_ip_address"),
                            )
                            continue

                        deleted_pip = await hub.exec.azurerm.network.public_ip_address.delete(
                            ctx,
                            pip_name,
                            pip_group,
                            azurerm_log_level="info",
                            **connection_auth,
                        )

        ret["result"] = True
        ret["comment"] = "Virtual machine {0} has been deleted.".format(name)
        ret["changes"] = {"old": vm, "new": {}}
        return ret

    ret["comment"] = "Failed to delete virtual machine {0}!".format(name)
    return ret



# -*- coding: utf-8 -*-
"""
Azure Resource Manager (ARM) Compute Availability Set State Module

.. versionadded:: 1.0.0

:maintainer: <devops@eitr.tech>
:maturity: new
:depends:
    * `azure <https://pypi.python.org/pypi/azure>`_ >= 4.0.0
    * `azure-common <https://pypi.python.org/pypi/azure-common>`_ >= 1.1.23
    * `azure-mgmt <https://pypi.python.org/pypi/azure-mgmt>`_ >= 4.0.0
    * `azure-mgmt-compute <https://pypi.python.org/pypi/azure-mgmt-compute>`_ >= 4.6.2
    * `azure-mgmt-network <https://pypi.python.org/pypi/azure-mgmt-network>`_ >= 4.0.0
    * `azure-mgmt-resource <https://pypi.python.org/pypi/azure-mgmt-resource>`_ >= 2.2.0
    * `azure-mgmt-storage <https://pypi.python.org/pypi/azure-mgmt-storage>`_ >= 2.0.0
    * `azure-mgmt-web <https://pypi.python.org/pypi/azure-mgmt-web>`_ >= 0.35.0
    * `azure-storage <https://pypi.python.org/pypi/azure-storage>`_ >= 0.36.0
    * `msrestazure <https://pypi.python.org/pypi/msrestazure>`_ >= 0.6.1
:platform: linux

:configuration: This module requires Azure Resource Manager credentials to be passed as a dictionary of
    keyword arguments to the ``connection_auth`` parameter in order to work properly. Since the authentication
    parameters are sensitive, it's recommended to pass them to the states via pillar.

    Required provider parameters:

    if using username and password:
      * ``subscription_id``
      * ``username``
      * ``password``

    if using a service principal:
      * ``subscription_id``
      * ``tenant``
      * ``client_id``
      * ``secret``

    Optional provider parameters:

    **cloud_environment**: Used to point the cloud driver to different API endpoints, such as Azure GovCloud. Possible values:
      * ``AZURE_PUBLIC_CLOUD`` (default)
      * ``AZURE_CHINA_CLOUD``
      * ``AZURE_US_GOV_CLOUD``
      * ``AZURE_GERMAN_CLOUD``

    Example Pillar for Azure Resource Manager authentication:

    .. code-block:: yaml

        azurerm:
            user_pass_auth:
                subscription_id: 3287abc8-f98a-c678-3bde-326766fd3617
                username: fletch
                password: 123pass
            mysubscription:
                subscription_id: 3287abc8-f98a-c678-3bde-326766fd3617
                tenant: ABCDEFAB-1234-ABCD-1234-ABCDEFABCDEF
                client_id: ABCDEFAB-1234-ABCD-1234-ABCDEFABCDEF
                secret: XXXXXXXXXXXXXXXXXXXXXXXX
                cloud_environment: AZURE_PUBLIC_CLOUD

    Example states using Azure Resource Manager authentication:

    .. code-block:: jinja

        Ensure availability set exists:
            azurerm.compute.availability_set.present:
                - name: my_avail_set
                - resource_group: my_rg
                - virtual_machines:
                    - my_vm1
                    - my_vm2
                - tags:
                    how_awesome: very
                    contact_name: Elmer Fudd Gantry
                - connection_auth: {{ profile }}

        Ensure availability set is absent:
            azurerm.compute.availability_set.absent:
                - name: other_avail_set
                - resource_group: my_rg
                - connection_auth: {{ profile }}

"""
# Python libs
from __future__ import absolute_import
import logging

log = logging.getLogger(__name__)

TREQ = {
    "present": {"require": ["states.azurerm.resource.group.present",]},
}


async def present(
    hub,
    ctx,
    name,
    resource_group,
    tags=None,
    platform_update_domain_count=None,
    platform_fault_domain_count=None,
    virtual_machines=None,
    sku=None,
    connection_auth=None,
    **kwargs,
):
    """
    .. versionadded:: 1.0.0

    Ensure an availability set exists.

    :param name:
        Name of the availability set.

    :param resource_group:
        The resource group assigned to the availability set.

    :param tags:
        A dictionary of strings can be passed as tag metadata to the availability set object.

    :param platform_update_domain_count:
        An optional parameter which indicates groups of virtual machines and underlying physical hardware that can be
        rebooted at the same time.

    :param platform_fault_domain_count:
        An optional parameter which defines the group of virtual machines that share a common power source and network
        switch.

    :param virtual_machines:
        A list of names of existing virtual machines to be included in the availability set.

    :param sku:
        The availability set SKU, which specifies whether the availability set is managed or not. Possible values are
        'Aligned' or 'Classic'. An 'Aligned' availability set is managed, 'Classic' is not.

    :param connection_auth:
        A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.

    Example usage:

    .. code-block:: yaml

        Ensure availability set exists:
            azurerm.compute.availability_set.present:
                - name: aset1
                - resource_group: group1
                - platform_update_domain_count: 5
                - platform_fault_domain_count: 3
                - sku: aligned
                - tags:
                    contact_name: Elmer Fudd Gantry
                - connection_auth: {{ profile }}

    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        if ctx["acct"]:
            connection_auth = ctx["acct"]
        else:
            ret[
                "comment"
            ] = "Connection information must be specified via acct or connection_auth dictionary!"
            return ret

    if sku:
        sku = {"name": sku.capitalize()}

    aset = await hub.exec.azurerm.compute.availability_set.get(
        ctx, name, resource_group, azurerm_log_level="info", **connection_auth
    )

    if "error" not in aset:
        tag_changes = await hub.exec.utils.dictdiffer.deep_diff(
            aset.get("tags", {}), tags or {}
        )
        if tag_changes:
            ret["changes"]["tags"] = tag_changes

        if platform_update_domain_count and (
            int(platform_update_domain_count)
            != aset.get("platform_update_domain_count")
        ):
            ret["changes"]["platform_update_domain_count"] = {
                "old": aset.get("platform_update_domain_count"),
                "new": platform_update_domain_count,
            }

        if platform_fault_domain_count and (
            int(platform_fault_domain_count) != aset.get("platform_fault_domain_count")
        ):
            ret["changes"]["platform_fault_domain_count"] = {
                "old": aset.get("platform_fault_domain_count"),
                "new": platform_fault_domain_count,
            }

        if sku and (sku["name"] != aset.get("sku", {}).get("name")):
            ret["changes"]["sku"] = {"old": aset.get("sku"), "new": sku}

        if virtual_machines:
            if not isinstance(virtual_machines, list):
                ret["comment"] = "Virtual machines must be supplied as a list!"
                return ret
            aset_vms = aset.get("virtual_machines", [])
            remote_vms = sorted(
                [vm["id"].split("/")[-1].lower() for vm in aset_vms if "id" in aset_vms]
            )
            local_vms = sorted([vm.lower() for vm in virtual_machines or []])
            if local_vms != remote_vms:
                ret["changes"]["virtual_machines"] = {
                    "old": aset_vms,
                    "new": virtual_machines,
                }

        if not ret["changes"]:
            ret["result"] = True
            ret["comment"] = "Availability set {0} is already present.".format(name)
            return ret

        if ctx["test"]:
            ret["result"] = None
            ret["comment"] = "Availability set {0} would be updated.".format(name)
            return ret

    else:
        ret["changes"] = {
            "old": {},
            "new": {
                "name": name,
                "virtual_machines": virtual_machines,
                "platform_update_domain_count": platform_update_domain_count,
                "platform_fault_domain_count": platform_fault_domain_count,
                "sku": sku,
                "tags": tags,
            },
        }

    if ctx["test"]:
        ret["comment"] = "Availability set {0} would be created.".format(name)
        ret["result"] = None
        return ret

    aset_kwargs = kwargs.copy()
    aset_kwargs.update(connection_auth)

    aset = await hub.exec.azurerm.compute.availability_set.create_or_update(
        ctx=ctx,
        name=name,
        resource_group=resource_group,
        virtual_machines=virtual_machines,
        platform_update_domain_count=platform_update_domain_count,
        platform_fault_domain_count=platform_fault_domain_count,
        sku=sku,
        tags=tags,
        **aset_kwargs,
    )

    if "error" not in aset:
        ret["result"] = True
        ret["comment"] = "Availability set {0} has been created.".format(name)
        return ret

    ret["comment"] = "Failed to create availability set {0}! ({1})".format(
        name, aset.get("error")
    )
    if not ret["result"]:
        ret["changes"] = {}
    return ret


async def absent(hub, ctx, name, resource_group, connection_auth=None, **kwargs):
    """
    .. versionadded:: 1.0.0

    Ensure an availability set does not exist in a resource group.

    :param name:
        Name of the availability set.

    :param resource_group:
        Name of the resource group containing the availability set.

    :param connection_auth:
        A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.

    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        if ctx["acct"]:
            connection_auth = ctx["acct"]
        else:
            ret[
                "comment"
            ] = "Connection information must be specified via acct or connection_auth dictionary!"
            return ret

    aset = await hub.exec.azurerm.compute.availability_set.get(
        ctx, name, resource_group, azurerm_log_level="info", **connection_auth
    )

    if "error" in aset:
        ret["result"] = True
        ret["comment"] = "Availability set {0} was not found.".format(name)
        return ret

    elif ctx["test"]:
        ret["comment"] = "Availability set {0} would be deleted.".format(name)
        ret["result"] = None
        ret["changes"] = {
            "old": aset,
            "new": {},
        }
        return ret

    deleted = await hub.exec.azurerm.compute.availability_set.delete(
        ctx, name, resource_group, **connection_auth
    )

    if deleted:
        ret["result"] = True
        ret["comment"] = "Availability set {0} has been deleted.".format(name)
        ret["changes"] = {"old": aset, "new": {}}
        return ret

    ret["comment"] = "Failed to delete availability set {0}!".format(name)
    return ret



# -*- coding: utf-8 -*-
"""
Azure Resource Manager (ARM) DNS Zone Execution Module

.. versionadded:: 1.0.0

:maintainer: <devops@eitr.tech>
:maturity: new
:depends:
    * `azure <https://pypi.python.org/pypi/azure>`_ >= 4.0.0
    * `azure-common <https://pypi.python.org/pypi/azure-common>`_ >= 1.1.23
    * `azure-mgmt <https://pypi.python.org/pypi/azure-mgmt>`_ >= 4.0.0
    * `azure-mgmt-compute <https://pypi.python.org/pypi/azure-mgmt-compute>`_ >= 4.6.2
    * `azure-mgmt-network <https://pypi.python.org/pypi/azure-mgmt-network>`_ >= 2.7.0
    * `azure-mgmt-resource <https://pypi.python.org/pypi/azure-mgmt-resource>`_ >= 2.2.0
    * `azure-mgmt-storage <https://pypi.python.org/pypi/azure-mgmt-storage>`_ >= 2.0.0
    * `azure-mgmt-web <https://pypi.python.org/pypi/azure-mgmt-web>`_ >= 0.35.0
    * `azure-storage <https://pypi.python.org/pypi/azure-storage>`_ >= 0.34.3
    * `msrestazure <https://pypi.python.org/pypi/msrestazure>`_ >= 0.6.2
:platform: linux

:configuration: This module requires Azure Resource Manager credentials to be passed as keyword arguments
to every function in order to work properly.

    Required provider parameters:

    if using username and password:
      * ``subscription_id``
      * ``username``
      * ``password``

    if using a service principal:
      * ``subscription_id``
      * ``tenant``
      * ``client_id``
      * ``secret``

    Optional provider parameters:

    **cloud_environment**: Used to point the cloud driver to different API endpoints, such as Azure GovCloud.
    Possible values:
      * ``AZURE_PUBLIC_CLOUD`` (default)
      * ``AZURE_CHINA_CLOUD``
      * ``AZURE_US_GOV_CLOUD``
      * ``AZURE_GERMAN_CLOUD``

"""

# Python libs
from __future__ import absolute_import
import logging

# Azure libs
HAS_LIBS = False
try:
    import azure.mgmt.dns.models  # pylint: disable=unused-import
    from msrest.exceptions import SerializationError
    from msrestazure.azure_exceptions import CloudError

    HAS_LIBS = True
except ImportError:
    pass

__func_alias__ = {"list_": "list"}

log = logging.getLogger(__name__)


async def create_or_update(hub, ctx, name, resource_group, **kwargs):
    """
    .. versionadded:: 1.0.0

    Creates or updates a DNS zone. Does not modify DNS records within the zone.

    :param name: The name of the DNS zone to create (without a terminating dot).

    :param resource_group: The name of the resource group.

    CLI Example:

    .. code-block:: bash

        azurerm.dns.zone.create_or_update myzone testgroup

    """
    # DNS zones are global objects
    kwargs["location"] = "global"

    dnsconn = await hub.exec.utils.azurerm.get_client(ctx, "dns", **kwargs)

    # Convert list of ID strings to list of dictionaries with id key.
    if isinstance(kwargs.get("registration_virtual_networks"), list):
        kwargs["registration_virtual_networks"] = [
            {"id": vnet} for vnet in kwargs["registration_virtual_networks"]
        ]

    if isinstance(kwargs.get("resolution_virtual_networks"), list):
        kwargs["resolution_virtual_networks"] = [
            {"id": vnet} for vnet in kwargs["resolution_virtual_networks"]
        ]

    try:
        zone_model = await hub.exec.utils.azurerm.create_object_model(
            "dns", "Zone", **kwargs
        )
    except TypeError as exc:
        result = {
            "error": "The object model could not be built. ({0})".format(str(exc))
        }
        return result

    try:
        zone = dnsconn.zones.create_or_update(
            zone_name=name,
            resource_group_name=resource_group,
            parameters=zone_model,
            if_match=kwargs.get("if_match"),
            if_none_match=kwargs.get("if_none_match"),
        )
        result = zone.as_dict()
    except CloudError as exc:
        await hub.exec.utils.azurerm.log_cloud_error("dns", str(exc), **kwargs)
        result = {"error": str(exc)}
    except SerializationError as exc:
        result = {
            "error": "The object model could not be parsed. ({0})".format(str(exc))
        }

    return result


async def delete(hub, ctx, name, resource_group, **kwargs):
    """
    .. versionadded:: 1.0.0

    Delete a DNS zone within a resource group.

    :param name: The name of the DNS zone to delete.

    :param resource_group: The name of the resource group.

    CLI Example:

    .. code-block:: bash

        azurerm.dns.zone.delete myzone testgroup

    """
    result = False
    dnsconn = await hub.exec.utils.azurerm.get_client(ctx, "dns", **kwargs)
    try:
        zone = dnsconn.zones.delete(
            zone_name=name,
            resource_group_name=resource_group,
            if_match=kwargs.get("if_match"),
        )
        zone.wait()
        result = True
    except CloudError as exc:
        await hub.exec.utils.azurerm.log_cloud_error("dns", str(exc), **kwargs)

    return result


async def get(hub, ctx, name, resource_group, **kwargs):
    """
    .. versionadded:: 1.0.0

    Get a dictionary representing a DNS zone's properties, but not the
    record sets within the zone.

    :param name: The DNS zone to get.

    :param resource_group: The name of the resource group.

    CLI Example:

    .. code-block:: bash

        azurerm.dns.zone.get myzone testgroup

    """
    dnsconn = await hub.exec.utils.azurerm.get_client(ctx, "dns", **kwargs)
    try:
        zone = dnsconn.zones.get(zone_name=name, resource_group_name=resource_group)
        result = zone.as_dict()

    except CloudError as exc:
        await hub.exec.utils.azurerm.log_cloud_error("dns", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


async def list_by_resource_group(hub, ctx, resource_group, top=None, **kwargs):
    """
    .. versionadded:: 1.0.0

    Lists the DNS zones in a resource group.

    :param resource_group: The name of the resource group.

    :param top: The maximum number of DNS zones to return. If not specified,
    returns up to 100 zones.

    CLI Example:

    .. code-block:: bash

        azurerm.dns.zone.list_by_resource_group testgroup

    """
    result = {}
    dnsconn = await hub.exec.utils.azurerm.get_client(ctx, "dns", **kwargs)
    try:
        zones = await hub.exec.utils.azurerm.paged_object_to_list(
            dnsconn.zones.list_by_resource_group(
                resource_group_name=resource_group, top=top
            )
        )

        for zone in zones:
            result[zone["name"]] = zone
    except CloudError as exc:
        await hub.exec.utils.azurerm.log_cloud_error("dns", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


async def list_(hub, ctx, top=None, **kwargs):
    """
    .. versionadded:: 1.0.0

    Lists the DNS zones in all resource groups in a subscription.

    :param top: The maximum number of DNS zones to return. If not specified,
    returns up to 100 zones.

    CLI Example:

    .. code-block:: bash

        azurerm.dns.zone.list

    """
    result = {}
    dnsconn = await hub.exec.utils.azurerm.get_client(ctx, "dns", **kwargs)
    try:
        zones = await hub.exec.utils.azurerm.paged_object_to_list(
            dnsconn.zones.list(top=top)
        )

        for zone in zones:
            result[zone["name"]] = zone
    except CloudError as exc:
        await hub.exec.utils.azurerm.log_cloud_error("dns", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result



# -*- coding: utf-8 -*-
"""
Azure Resource Manager (ARM) Diagnostic Setting Execution Module

.. versionadded:: 1.0.0

:maintainer: <devops@eitr.tech>
:maturity: new
:depends:
    * `azure <https://pypi.python.org/pypi/azure>`_ >= 4.0.0
    * `azure-common <https://pypi.python.org/pypi/azure-common>`_ >= 1.1.23
    * `azure-mgmt <https://pypi.python.org/pypi/azure-mgmt>`_ >= 4.0.0
    * `azure-mgmt-compute <https://pypi.python.org/pypi/azure-mgmt-compute>`_ >= 4.6.2
    * `azure-mgmt-monitor <https://pypi.org/project/azure-mgmt-monitor>`_ >= 0.5.2
    * `azure-mgmt-network <https://pypi.python.org/pypi/azure-mgmt-network>`_ >= 2.7.0
    * `azure-mgmt-resource <https://pypi.python.org/pypi/azure-mgmt-resource>`_ >= 2.2.0
    * `azure-mgmt-storage <https://pypi.python.org/pypi/azure-mgmt-storage>`_ >= 2.0.0
    * `azure-mgmt-web <https://pypi.python.org/pypi/azure-mgmt-web>`_ >= 0.35.0
    * `azure-storage <https://pypi.python.org/pypi/azure-storage>`_ >= 0.34.3
    * `msrestazure <https://pypi.python.org/pypi/msrestazure>`_ >= 0.6.2
:platform: linux

:configuration: This module requires Azure Resource Manager credentials to be passed as keyword arguments
to every function in order to work properly.

    Required provider parameters:

    if using username and password:
      * ``subscription_id``
      * ``username``
      * ``password``

    if using a service principal:
      * ``subscription_id``
      * ``tenant``
      * ``client_id``
      * ``secret``

    Optional provider parameters:

    **cloud_environment**: Used to point the cloud driver to different API endpoints, such as Azure GovCloud.
    Possible values:
      * ``AZURE_PUBLIC_CLOUD`` (default)
      * ``AZURE_CHINA_CLOUD``
      * ``AZURE_US_GOV_CLOUD``
      * ``AZURE_GERMAN_CLOUD``

"""

# Python libs
from __future__ import absolute_import
import logging

# Azure libs
HAS_LIBS = False
try:
    import azure.mgmt.monitor.models  # pylint: disable=unused-import
    from msrest.exceptions import SerializationError
    from msrestazure.azure_exceptions import CloudError
    from azure.mgmt.monitor.models import ErrorResponseException

    HAS_LIBS = True
except ImportError:
    pass

__func_alias__ = {"list_": "list"}

log = logging.getLogger(__name__)


async def create_or_update(
    hub,
    ctx,
    name,
    resource_uri,
    metrics,
    logs,
    workspace_id=None,
    storage_account_id=None,
    service_bus_rule_id=None,
    event_hub_authorization_rule_id=None,
    event_hub_name=None,
    **kwargs,
):
    """
    .. versionadded:: 1.0.0

    Create or update diagnostic settings for the specified resource. At least one destination for the diagnostic
        setting logs is required. Any combination of the following destinations is acceptable:
            1. Archive the diagnostic settings to a stroage account. This would require the storage_account_id param.
            2. Stream the diagnostic settings to an event hub. This would require the event_hub_name and
               event_hub_authorization_rule_id params.
            3. Send the diagnostic settings to Log Analytics. This would require the workspace_id param.

    :param name: The name of the diagnostic setting.

    :param resource_uri: The identifier of the resource.

    :param metrics: A list of dictionaries representing valid MetricSettings objects. If this list is empty, then the
        list passed as the logs parameter must have at least one element. Valid parameters are:
        - ``category``: Name of a diagnostic metric category for the resource type this setting is applied to. To obtain
          the list of diagnostic metric categories for a resource, first perform a GET diagnostic setting operation.
          This is a required parameter.
        - ``enabled``: A value indicating whether this category is enabled. This is a required parameter.
        - ``time_grain``: An optional timegrain of the metric in ISO-8601 format.
        - ``retention_policy``: An optional dictionary representing a RetentionPolicy object for the specified category.
          The default retention policy for a diagnostic setting is {'enabled': False, 'days': 0}. Required parameters
          include:
            - ``days``: The number of days for the retention in days. A value of 0 will retain the events indefinitely.
            - ``enabled``: A value indicating whether the retention policy is enabled.

    :param logs: A list of dictionaries representing valid LogSettings objects. If this list is empty, then the list
        passed as the metrics parameter must have at least one element. Valid parameters are:
        - ``category``: Name of a diagnostic log category for the resource type this setting is applied to. To obtain
          the list of diagnostic log categories for a resource, first perform a GET diagnostic setting operation.
          This is a required parameter.
        - ``enabled``: A value indicating whether this category is enabled. This is a required parameter.
        - ``retention_policy``: An optional dictionary representing a RetentionPolicy object for the specified category.
          The default retention policy for a diagnostic setting is {'enabled': False, 'days': 0}. Required parameters
          include:
            - ``days``: The number of days for the retention in days. A value of 0 will retain the events indefinitely.
            - ``enabled``: A value indicating whether the retention policy is enabled.

    :param workspace_id: The workspace (resource) ID for the Log Analytics workspace to which you would like to
        send Diagnostic Logs.

    :param storage_account_id: The resource ID of the storage account to which you would like to send Diagnostic Logs.

    :param service_bus_rule_id: The service bus rule ID of the diagnostic setting. This is here to
        maintain backwards compatibility.

    :param event_hub_authorization_rule_id: The resource ID for the event hub authorization rule.

    :param event_hub_name: The name of the event hub. If none is specified, the default event hub will be selected.

    CLI Example:

    .. code-block:: bash

        azurerm.monitor.diagnostic_setting.create_or_update test_name test_uri test_metrics test_logs \
                  test_destination

    """
    result = {}
    moniconn = await hub.exec.utils.azurerm.get_client(ctx, "monitor", **kwargs)

    try:
        diagmodel = await hub.exec.utils.azurerm.create_object_model(
            "monitor",
            "DiagnosticSettingsResource",
            metrics=metrics,
            logs=logs,
            workspace_id=workspace_id,
            storage_account_id=storage_account_id,
            service_bus_rule_id=service_bus_rule_id,
            event_hub_authorization_rule_id=event_hub_authorization_rule_id,
            event_hub_name=event_hub_name,
            **kwargs,
        )
    except TypeError as exc:
        result = {
            "error": "The object model could not be built. ({0})".format(str(exc))
        }
        return result

    try:
        diag = moniconn.diagnostic_settings.create_or_update(
            name=name, resource_uri=resource_uri, parameters=diagmodel
        )

        result = diag.as_dict()
    except (CloudError, ErrorResponseException) as exc:
        await hub.exec.utils.azurerm.log_cloud_error("monitor", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


async def delete(hub, ctx, name, resource_uri, **kwargs):
    """
    .. versionadded:: 1.0.0

    Deletes existing diagnostic settings for the specified resource.

    :param name: The name of the diagnostic setting.

    :param resource_uri: The identifier of the resource.

    CLI Example:

    .. code-block:: bash

        azurerm.monitor.diagnostic_setting.delete test_name test_uri

    """
    result = False
    moniconn = await hub.exec.utils.azurerm.get_client(ctx, "monitor", **kwargs)
    try:
        diag = moniconn.diagnostic_settings.delete(
            name=name, resource_uri=resource_uri, **kwargs
        )

        result = True
    except CloudError as exc:
        await hub.exec.utils.azurerm.log_cloud_error("monitor", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


async def get(hub, ctx, name, resource_uri, **kwargs):
    """
    .. versionadded:: 1.0.0

    Gets the active diagnostic settings for the specified resource.

    :param name: The name of the diagnostic setting.

    :param resource_uri: The identifier of the resource.

    CLI Example:

    .. code-block:: bash

        azurerm.monitor.diagnostic_setting.get test_name test_uri

    """
    result = {}
    moniconn = await hub.exec.utils.azurerm.get_client(ctx, "monitor", **kwargs)

    try:
        diag = moniconn.diagnostic_settings.get(
            name=name, resource_uri=resource_uri, **kwargs
        )

        result = diag.as_dict()
    except (CloudError, ErrorResponseException) as exc:
        await hub.exec.utils.azurerm.log_cloud_error("monitor", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


async def list_(hub, ctx, resource_uri, **kwargs):
    """
    .. versionadded:: 1.0.0

    Gets the active diagnostic settings list for the specified resource.

    :param resource_uri: The identifier of the resource.

    CLI Example:

    .. code-block:: bash

        azurerm.monitor.diagnostic_setting.list test_uri

    """
    result = {}
    moniconn = await hub.exec.utils.azurerm.get_client(ctx, "monitor", **kwargs)

    try:
        diag = moniconn.diagnostic_settings.list(resource_uri=resource_uri, **kwargs)

        values = diag.as_dict().get("value", [])
        for value in values:
            result[value["name"]] = value
    except CloudError as exc:
        await hub.exec.utils.azurerm.log_cloud_error("monitor", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result



# -*- coding: utf-8 -*-
"""
Azure Resource Manager (ARM) Management Lock State Module

.. versionadded:: 2.0.0

:maintainer: <devops@eitr.tech>
:maturity: new
:depends:
    * `azure <https://pypi.python.org/pypi/azure>`_ >= 4.0.0
    * `azure-common <https://pypi.python.org/pypi/azure-common>`_ >= 1.1.23
    * `azure-mgmt <https://pypi.python.org/pypi/azure-mgmt>`_ >= 4.0.0
    * `azure-mgmt-compute <https://pypi.python.org/pypi/azure-mgmt-compute>`_ >= 4.6.2
    * `azure-mgmt-network <https://pypi.python.org/pypi/azure-mgmt-network>`_ >= 2.7.0
    * `azure-mgmt-resource <https://pypi.python.org/pypi/azure-mgmt-resource>`_ >= 2.2.0
    * `azure-mgmt-storage <https://pypi.python.org/pypi/azure-mgmt-storage>`_ >= 2.0.0
    * `azure-mgmt-web <https://pypi.python.org/pypi/azure-mgmt-web>`_ >= 0.35.0
    * `azure-storage <https://pypi.python.org/pypi/azure-storage>`_ >= 0.34.3
    * `msrestazure <https://pypi.python.org/pypi/msrestazure>`_ >= 0.6.2
:platform: linux

:configuration: This module requires Azure Resource Manager credentials to be passed as a dictionary of
    keyword arguments to the ``connection_auth`` parameter in order to work properly. Since the authentication
    parameters are sensitive, it's recommended to pass them to the states via pillar.

    Required provider parameters:

    if using username and password:
      * ``subscription_id``
      * ``username``
      * ``password``

    if using a service principal:
      * ``subscription_id``
      * ``tenant``
      * ``client_id``
      * ``secret``

    Optional provider parameters:

    **cloud_environment**: Used to point the cloud driver to different API endpoints, such as Azure GovCloud. Possible values:
      * ``AZURE_PUBLIC_CLOUD`` (default)
      * ``AZURE_CHINA_CLOUD``
      * ``AZURE_US_GOV_CLOUD``
      * ``AZURE_GERMAN_CLOUD``

    Example configuration for Azure Resource Manager authentication:

    .. code-block:: yaml

        azurerm:
            user_pass_auth:
                subscription_id: 3287abc8-f98a-c678-3bde-326766fd3617
                username: fletch
                password: 123pass
            mysubscription:
                subscription_id: 3287abc8-f98a-c678-3bde-326766fd3617
                tenant: ABCDEFAB-1234-ABCD-1234-ABCDEFABCDEF
                client_id: ABCDEFAB-1234-ABCD-1234-ABCDEFABCDEF
                secret: XXXXXXXXXXXXXXXXXXXXXXXX
                cloud_environment: AZURE_PUBLIC_CLOUD

"""
# Import Python libs
from __future__ import absolute_import
import logging

log = logging.getLogger(__name__)


async def present_by_scope(
    hub,
    ctx,
    name,
    scope,
    lock_level,
    notes=None,
    owners=None,
    connection_auth=None,
    **kwargs,
):
    """
    .. versionadded:: 2.0.0

    Ensure a management lock exists by scope.

    :param name: The name of the lock. The lock name can be a maximum of 260 characters. It cannot contain <, > %, &,
        :, ?, /, or any control characters.

    :param scope: The scope for the lock. When providing a scope for the assignment,
        use '/subscriptions/{subscriptionId}' for subscriptions,
        '/subscriptions/{subscriptionId}/resourcegroups/{resourceGroupName}' for resource groups, and
        '/subscriptions/{subscriptionId}/resourcegroups/{resourceGroupName}/providers/{resourceProviderNamespace}/{parentResourcePathIfPresent}/{resourceType}/{resourceName}'
        for resources.

    :param lock_level: The level of the lock. Possible values are: 'CanNotDelete' and 'ReadOnly'. CanNotDelete means
        authorized users are able to read and modify the resources, but not delete. ReadOnly means authorized users
        can only read from a resource, but they can't modify or delete it.

    :param notes: An optional string representing notes about the lock. Maximum of 512 characters.

    :param owners: An optional list of strings representing owners of the lock. Each string represents the application
        id of the lock owner.

    :param connection_auth: A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.

    Example usage:

    .. code-block:: yaml

        Ensure management lock exists by scope:
            azurerm.resource.management_lock.present_by_scope:
                - name: my_lock
                - scope: my_scope
                - lock_level: 'ReadOnly'
                - connection_auth: {{ profile }}

    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        if ctx["acct"]:
            connection_auth = ctx["acct"]
        else:
            ret[
                "comment"
            ] = "Connection information must be specified via acct or connection_auth dictionary!"
            return ret

    lock = await hub.exec.azurerm.resource.management_lock.get_by_scope(
        ctx, name, scope, azurerm_log_level="info", **connection_auth
    )

    if "error" not in lock:
        if lock_level != lock.get("level"):
            ret["changes"]["level"] = {"old": lock.get("level"), "new": lock_level}

        if notes != lock.get("notes"):
            ret["changes"]["notes"] = {"old": lock.get("notes"), "new": notes}

        if owners:
            new_owners = owners.sort()
            lock_owners = lock.get("owners", [])
            if lock_owners:
                # Extracts the application_id value from each dictionary that represents a ManagementLockOwner object
                old_owners = [owner.get("application_id") for owner in lock_owners]
                old_owners = old_owners.sort()
            else:
                old_owners = []

            if old_owners != new_owners:
                ret["changes"]["owners"] = {"old": old_owners, "new": new_owners}

        if not ret["changes"]:
            ret["result"] = True
            ret["comment"] = "Management lock {0} is already present.".format(name)
            return ret

        if ctx["test"]:
            ret["result"] = None
            ret["comment"] = "Management lock {0} would be updated.".format(name)
            return ret

    else:
        ret["changes"] = {
            "old": {},
            "new": {"name": name, "scope": scope, "lock_level": lock_level,},
        }

        if owners:
            ret["changes"]["new"]["owners"] = owners
        if notes:
            ret["changes"]["new"]["notes"] = notes

    if ctx["test"]:
        ret["comment"] = "Management lock {0} would be created.".format(name)
        ret["result"] = None
        return ret

    lock_kwargs = kwargs.copy()
    lock_kwargs.update(connection_auth)

    lock = await hub.exec.azurerm.resource.management_lock.create_or_update_by_scope(
        ctx=ctx,
        name=name,
        scope=scope,
        lock_level=lock_level,
        notes=notes,
        owners=owners,
        **lock_kwargs,
    )

    if "error" not in lock:
        ret["result"] = True
        ret["comment"] = "Management lock {0} has been created.".format(name)
        return ret

    ret["comment"] = "Failed to create management lock {0}! ({1})".format(
        name, lock.get("error")
    )
    if not ret["result"]:
        ret["changes"] = {}
    return ret


async def absent_by_scope(hub, ctx, name, scope, connection_auth=None, **kwargs):
    """
    .. versionadded:: 2.0.0

    Ensure a management lock does not exist by scope.

    :param name: The name of the lock. The lock name can be a maximum of 260 characters. It cannot contain <, > %, &,
        :, ?, /, or any control characters.

    :param scope: The scope for the lock. When providing a scope for the assignment,
        use '/subscriptions/{subscriptionId}' for subscriptions,
        '/subscriptions/{subscriptionId}/resourcegroups/{resourceGroupName}' for resource groups, and
        '/subscriptions/{subscriptionId}/resourcegroups/{resourceGroupName}/providers/{resourceProviderNamespace}/{parentResourcePathIfPresent}/{resourceType}/{resourceName}' for resources.

    :param connection_auth: A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.

    Example usage:

    .. code-block:: yaml

        Ensure management lock absent by scope:
            azurerm.resource.management_lock.absent_by_scope:
                - name: my_lock
                - scope: my_scope
                - connection_auth: {{ profile }}

    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        if ctx["acct"]:
            connection_auth = ctx["acct"]
        else:
            ret[
                "comment"
            ] = "Connection information must be specified via acct or connection_auth dictionary!"
            return ret

    lock = await hub.exec.azurerm.resource.management_lock.get_by_scope(
        ctx, name, scope, azurerm_log_level="info", **connection_auth
    )

    if "error" in lock:
        ret["result"] = True
        ret["comment"] = "Management lock {0} was not found.".format(name)
        return ret

    elif ctx["test"]:
        ret["comment"] = "Management lock {0} would be deleted.".format(name)
        ret["result"] = None
        ret["changes"] = {
            "old": lock,
            "new": {},
        }
        return ret

    deleted = await hub.exec.azurerm.resource.management_lock.delete_by_scope(
        ctx, name, scope, **connection_auth
    )

    if deleted:
        ret["result"] = True
        ret["comment"] = "Management lock {0} has been deleted.".format(name)
        ret["changes"] = {"old": lock, "new": {}}
        return ret

    ret["comment"] = "Failed to delete management lock {0}!".format(name)
    return ret


async def present_at_resource_level(
    hub,
    ctx,
    name,
    lock_level,
    resource_group,
    resource,
    resource_type,
    resource_provider_namespace,
    parent_resource_path=None,
    notes=None,
    owners=None,
    connection_auth=None,
    **kwargs,
):
    """
    .. versionadded:: 2.0.0

    Ensure a management lock exists at resource level.

    :param name: The name of the lock. The lock name can be a maximum of 260 characters. It cannot contain<, > %, &,
        :, ?, /, or any control characters.

    :param lock_level: The level of the lock. Possible values are: 'CanNotDelete' and 'ReadOnly'. CanNotDelete means
        authorized users are able to read and modify the resources, but not delete. ReadOnly means authorized users
        can only read from a resource, but they can't modify or delete it.

    :param resource_group: The name of the resource group containing the resource to lock.

    :param resource: The name of the resource to lock.

    :param resource_type: The resource type of the resource to lock.

    :param resource_provider_namespace: The resource provider namespace of the resource to lock.

    :param parent_resource_path: The parent resource identity.

    :param notes: An optional string representing notes about the lock. Maximum of 512 characters.

    :param owners: An optional list of strings representing owners of the lock. Each string represents the application
        id of the lock owner.

    :param connection_auth: A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.

    Example usage:

    .. code-block:: yaml

        Ensure management lock exists at resource level:
            azurerm.resource.management_lock.present_at_resource_level:
                - name: my_lock
                - resource_group: my_rg
                - resource: my_resource
                - resource_type: my_type
                - resource_provider_namespace: my_namespace
                - lock_level: 'ReadOnly'
                - connection_auth: {{ profile }}

    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        if ctx["acct"]:
            connection_auth = ctx["acct"]
        else:
            ret[
                "comment"
            ] = "Connection information must be specified via acct or connection_auth dictionary!"
            return ret

    lock = await hub.exec.azurerm.resource.management_lock.get_at_resource_level(
        ctx,
        name,
        resource_group,
        resource,
        resource_type,
        resource_provider_namespace,
        parent_resource_path,
        azurerm_log_level="info",
        **connection_auth,
    )

    if "error" not in lock:
        if lock_level != lock.get("level"):
            ret["changes"]["level"] = {"old": lock.get("level"), "new": lock_level}

        if notes != lock.get("notes"):
            ret["changes"]["notes"] = {"old": lock.get("notes"), "new": notes}

        if owners:
            new_owners = owners.sort()
            lock_owners = lock.get("owners", [])
            if lock_owners:
                # Extracts the application_id value from each dictionary that represents a ManagementLockOwner object
                old_owners = [owner.get("application_id") for owner in lock_owners]
                old_owners = old_owners.sort()
            else:
                old_owners = []

            if old_owners != new_owners:
                ret["changes"]["owners"] = {"old": old_owners, "new": new_owners}

        if not ret["changes"]:
            ret["result"] = True
            ret["comment"] = "Management lock {0} is already present.".format(name)
            return ret

        if ctx["test"]:
            ret["result"] = None
            ret["comment"] = "Management lock {0} would be updated.".format(name)
            return ret

    else:
        ret["changes"] = {
            "old": {},
            "new": {
                "name": name,
                "resource_group": resource_group,
                "lock_level": lock_level,
                "resource": resource,
                "resource_type": resource_type,
                "resource_provider_namespace": resource_provider_namespace,
            },
        }

        if owners:
            ret["changes"]["new"]["owners"] = owners
        if notes:
            ret["changes"]["new"]["notes"] = notes
        if parent_resource_path:
            ret["changes"]["new"]["parent_resource_path"] = parent_resource_path

    if ctx["test"]:
        ret["comment"] = "Management lock {0} would be created.".format(name)
        ret["result"] = None
        return ret

    lock_kwargs = kwargs.copy()
    lock_kwargs.update(connection_auth)

    lock = await hub.exec.azurerm.resource.management_lock.create_or_update_at_resource_level(
        ctx=ctx,
        name=name,
        resource_group=resource_group,
        resource=resource,
        resource_type=resource_type,
        resource_provider_namespace=resource_provider_namespace,
        parent_resource_path=parent_resource_path,
        lock_level=lock_level,
        notes=notes,
        owners=owners,
        **lock_kwargs,
    )

    if "error" not in lock:
        ret["result"] = True
        ret["comment"] = "Management lock {0} has been created.".format(name)
        return ret

    ret["comment"] = "Failed to create management lock {0}! ({1})".format(
        name, lock.get("error")
    )
    if not ret["result"]:
        ret["changes"] = {}
    return ret


async def absent_at_resource_level(
    hub,
    ctx,
    name,
    resource_group,
    resource,
    resource_type,
    resource_provider_namespace,
    parent_resource_path=None,
    connection_auth=None,
    **kwargs,
):
    """
    .. versionadded:: 2.0.0

    Ensure a management lock does not exist at the resource level.

    :param name: The name of the lock. The lock name can be a maximum of 260 characters. It cannot contain <, > %, &,
        :, ?, /, or any control characters.

    :param resource_group: The name of the resource group containing the resource with the lock to delete.

    :param resource: The name of the resource with the lock to delete.

    :param resource_type: The resource type of the resource with the lock to delete.

    :param resource_provider_namespace: The resource provider namespace of the resource with the lock to delete.

    :param parent_resource_path: The parent resource identity.

    :param connection_auth: A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.

    Example usage:

    .. code-block:: yaml

        Ensure management lock absent at resource level:
            azurerm.resource.management_lock.absent_at_resource_level:
                - name: my_lock
                - resource_group: my_rg
                - resource: my_resource
                - resource_type: my_type
                - resource_provider_namespace: my_namespace
                - connection_auth: {{ profile }}

    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        if ctx["acct"]:
            connection_auth = ctx["acct"]
        else:
            ret[
                "comment"
            ] = "Connection information must be specified via acct or connection_auth dictionary!"
            return ret

    lock = await hub.exec.azurerm.resource.management_lock.get_at_resource_level(
        ctx,
        name,
        resource_group,
        resource,
        resource_type,
        resource_provider_namespace,
        parent_resource_path=parent_resource_path,
        azurerm_log_level="info",
        **connection_auth,
    )

    if "error" in lock:
        ret["result"] = True
        ret["comment"] = "Management lock {0} was not found.".format(name)
        return ret

    elif ctx["test"]:
        ret["comment"] = "Management lock {0} would be deleted.".format(name)
        ret["result"] = None
        ret["changes"] = {
            "old": lock,
            "new": {},
        }
        return ret

    deleted = await hub.exec.azurerm.resource.management_lock.delete_at_resource_level(
        ctx,
        name,
        resource_group,
        resource,
        resource_type,
        resource_provider_namespace,
        parent_resource_path=parent_resource_path,
        **connection_auth,
    )

    if deleted:
        ret["result"] = True
        ret["comment"] = "Management lock {0} has been deleted.".format(name)
        ret["changes"] = {"old": lock, "new": {}}
        return ret

    ret["comment"] = "Failed to delete management lock {0}!".format(name)
    return ret


async def present(
    hub,
    ctx,
    name,
    lock_level,
    resource_group=None,
    notes=None,
    owners=None,
    connection_auth=None,
    **kwargs,
):
    """
    .. versionadded:: 2.0.0

    Ensure a management lock exists. By default this module ensures that the management lock exists at the
        subscription level. If you would like to ensure that the management lock exists at the resource group level
        instead, you can specify a resource group using the resource_group parameter.

    :param name: The name of the lock. The lock name can be a maximum of 260 characters. It cannot contain <, > %, &,
        :, ?, /, or any control characters.

    :param lock_level: The level of the lock. Possible values are: 'CanNotDelete' and 'ReadOnly'. CanNotDelete means
        authorized users are able to read and modify the resources, but not delete. ReadOnly means authorized users
        can only read from a resource, but they can't modify or delete it.

    :param resource_group: The name of the resource group. This is an optional parameter.

    :param notes: An optional string representing notes about the lock. Maximum of 512 characters.

    :param owners: An optional list of strings representing owners of the lock. Each string represents the application
        id of the lock owner.

    :param connection_auth: A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.

    Example usage:

    .. code-block:: yaml

        Ensure management lock exists:
            azurerm.resource.management_lock.present:
                - name: my_lock
                - lock_level: 'ReadOnly'
                - connection_auth: {{ profile }}

    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        if ctx["acct"]:
            connection_auth = ctx["acct"]
        else:
            ret[
                "comment"
            ] = "Connection information must be specified via acct or connection_auth dictionary!"
            return ret

    if resource_group:
        lock = await hub.exec.azurerm.resource.management_lock.get_at_resource_group_level(
            ctx, name, resource_group, azurerm_log_level="info", **connection_auth
        )
    else:
        lock = await hub.exec.azurerm.resource.management_lock.get_at_subscription_level(
            ctx, name, azurerm_log_level="info", **connection_auth
        )

    if "error" not in lock:
        if lock_level != lock.get("level"):
            ret["changes"]["level"] = {"old": lock.get("level"), "new": lock_level}

        if notes != lock.get("notes"):
            ret["changes"]["notes"] = {"old": lock.get("notes"), "new": notes}

        if owners:
            new_owners = owners.sort()
            lock_owners = lock.get("owners", [])
            if lock_owners:
                # Extracts the application_id value from each dictionary that represents a ManagementLockOwner object
                old_owners = [owner.get("application_id") for owner in lock_owners]
                old_owners = old_owners.sort()
            else:
                old_owners = []

            if old_owners != new_owners:
                ret["changes"]["owners"] = {"old": old_owners, "new": new_owners}

        if not ret["changes"]:
            ret["result"] = True
            ret["comment"] = "Management lock {0} is already present.".format(name)
            return ret

        if ctx["test"]:
            ret["result"] = None
            ret["comment"] = "Management lock {0} would be updated.".format(name)
            return ret

    else:
        ret["changes"] = {"old": {}, "new": {"name": name, "lock_level": lock_level,}}

        if resource_group:
            ret["changes"]["new"]["resource_group"] = resource_group
        if owners:
            ret["changes"]["new"]["owners"] = owners
        if notes:
            ret["changes"]["new"]["notes"] = notes

    if ctx["test"]:
        ret["comment"] = "Management lock {0} would be created.".format(name)
        ret["result"] = None
        return ret

    lock_kwargs = kwargs.copy()
    lock_kwargs.update(connection_auth)

    if resource_group:
        lock = await hub.exec.azurerm.resource.management_lock.create_or_update_at_resource_group_level(
            ctx=ctx,
            name=name,
            resource_group=resource_group,
            lock_level=lock_level,
            notes=notes,
            owners=owners,
            **lock_kwargs,
        )
    else:
        lock = await hub.exec.azurerm.resource.management_lock.create_or_update_at_subscription_level(
            ctx=ctx,
            name=name,
            lock_level=lock_level,
            notes=notes,
            owners=owners,
            **lock_kwargs,
        )

    if "error" not in lock:
        ret["result"] = True
        ret["comment"] = "Management lock {0} has been created.".format(name)
        return ret

    ret["comment"] = "Failed to create management lock {0}! ({1})".format(
        name, lock.get("error")
    )
    if not ret["result"]:
        ret["changes"] = {}
    return ret


async def absent(hub, ctx, name, resource_group=None, connection_auth=None, **kwargs):
    """
    .. versionadded:: 2.0.0

    Ensure a management lock does not exist. By default this module ensures that the management lock does not exist at
        the subscription level. If you would like to ensure that the management lock does not exist at the resource
        group level instead, you can specify a resource group using the resource_group parameter.

    :param name: The name of the lock. The lock name can be a maximum of 260 characters. It cannot contain <, > %, &,
        :, ?, /, or any control characters.

    :param resource_group: The name of the resource group. This is an optional parameter.

    :param connection_auth: A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.

    Example usage:

    .. code-block:: yaml

        Ensure management lock is absent:
            azurerm.resource.management_lock.absent:
                - name: my_lock
                - connection_auth: {{ profile }}

    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        if ctx["acct"]:
            connection_auth = ctx["acct"]
        else:
            ret[
                "comment"
            ] = "Connection information must be specified via acct or connection_auth dictionary!"
            return ret

    if resource_group:
        lock = await hub.exec.azurerm.resource.management_lock.get_at_resource_group_level(
            ctx, name, resource_group, azurerm_log_level="info", **connection_auth
        )
    else:
        lock = await hub.exec.azurerm.resource.management_lock.get_at_subscription_level(
            ctx, name, azurerm_log_level="info", **connection_auth
        )

    if "error" in lock:
        ret["result"] = True
        ret["comment"] = "Management lock {0} was not found.".format(name)
        return ret

    elif ctx["test"]:
        ret["comment"] = "Management lock {0} would be deleted.".format(name)
        ret["result"] = None
        ret["changes"] = {
            "old": lock,
            "new": {},
        }
        return ret

    if resource_group:
        deleted = await hub.exec.azurerm.resource.management_lock.delete_at_resource_group_level(
            ctx, name, resource_group, **connection_auth
        )
    else:
        deleted = await hub.exec.azurerm.resource.management_lock.delete_at_subscription_level(
            ctx, name, **connection_auth
        )

    if deleted:
        ret["result"] = True
        ret["comment"] = "Management lock {0} has been deleted.".format(name)
        ret["changes"] = {"old": lock, "new": {}}
        return ret

    ret["comment"] = "Failed to delete management lock {0}!".format(name)
    return ret



# -*- coding: utf-8 -*-
"""
Azure Resource Manager (ARM) Resource Policy Execution Module

.. versionadded:: 1.0.0

:maintainer: <devops@eitr.tech>
:maturity: new
:depends:
    * `azure <https://pypi.python.org/pypi/azure>`_ >= 4.0.0
    * `azure-common <https://pypi.python.org/pypi/azure-common>`_ >= 1.1.23
    * `azure-mgmt <https://pypi.python.org/pypi/azure-mgmt>`_ >= 4.0.0
    * `azure-mgmt-compute <https://pypi.python.org/pypi/azure-mgmt-compute>`_ >= 4.6.2
    * `azure-mgmt-network <https://pypi.python.org/pypi/azure-mgmt-network>`_ >= 2.7.0
    * `azure-mgmt-resource <https://pypi.python.org/pypi/azure-mgmt-resource>`_ >= 2.2.0
    * `azure-mgmt-storage <https://pypi.python.org/pypi/azure-mgmt-storage>`_ >= 2.0.0
    * `azure-mgmt-web <https://pypi.python.org/pypi/azure-mgmt-web>`_ >= 0.35.0
    * `azure-storage <https://pypi.python.org/pypi/azure-storage>`_ >= 0.34.3
    * `msrestazure <https://pypi.python.org/pypi/msrestazure>`_ >= 0.6.2
:platform: linux

:configuration: This module requires Azure Resource Manager credentials to be passed as keyword arguments
    to every function in order to work properly.

    Required provider parameters:

    if using username and password:
      * ``subscription_id``
      * ``username``
      * ``password``

    if using a service principal:
      * ``subscription_id``
      * ``tenant``
      * ``client_id``
      * ``secret``

    Optional provider parameters:

**cloud_environment**: Used to point the cloud driver to different API endpoints, such as Azure GovCloud.
    Possible values:
      * ``AZURE_PUBLIC_CLOUD`` (default)
      * ``AZURE_CHINA_CLOUD``
      * ``AZURE_US_GOV_CLOUD``
      * ``AZURE_GERMAN_CLOUD``

"""

# Python libs
from __future__ import absolute_import
from json import loads, dumps
import logging

# Azure libs
HAS_LIBS = False
try:
    import azure.mgmt.resource.resources.models  # pylint: disable=unused-import
    from msrest.exceptions import SerializationError
    from msrestazure.azure_exceptions import CloudError

    HAS_LIBS = True
except ImportError:
    pass

log = logging.getLogger(__name__)


async def assignment_delete(hub, ctx, name, scope, **kwargs):
    """
    .. versionadded:: 1.0.0

    Delete a policy assignment.

    :param name: The name of the policy assignment to delete.

    :param scope: The scope of the policy assignment.

    CLI Example:

    .. code-block:: bash

        azurerm.resource.policy.assignment_delete testassign \
        /subscriptions/bc75htn-a0fhsi-349b-56gh-4fghti-f84852

    """
    result = False
    polconn = await hub.exec.utils.azurerm.get_client(ctx, "policy", **kwargs)
    try:
        # pylint: disable=unused-variable
        policy = polconn.policy_assignments.delete(
            policy_assignment_name=name, scope=scope
        )
        result = True
    except CloudError as exc:
        await hub.exec.utils.azurerm.log_cloud_error("resource", str(exc), **kwargs)

    return result


async def assignment_create(hub, ctx, name, scope, definition_name, **kwargs):
    """
    .. versionadded:: 1.0.0

    Create a policy assignment.

    :param name: The name of the policy assignment to create.

    :param scope: The scope of the policy assignment.

    :param definition_name: The name of the policy definition to assign.

    CLI Example:

    .. code-block:: bash

        azurerm.resource.policy.assignment_create testassign \
        /subscriptions/bc75htn-a0fhsi-349b-56gh-4fghti-f84852 testpolicy

    """
    polconn = await hub.exec.utils.azurerm.get_client(ctx, "policy", **kwargs)

    # "get" doesn't work for built-in policies per https://github.com/Azure/azure-cli/issues/692
    # Uncomment this section when the ticket above is resolved.
    #  BEGIN
    # definition = definition_get(
    #     name=definition_name,
    #     **kwargs
    # )
    #  END

    # Delete this section when the ticket above is resolved.
    #  BEGIN
    definition_list = await hub.exec.azurerm.resource.policy.definitions_list(
        ctx=ctx, **kwargs
    )
    if definition_name in definition_list:
        definition = definition_list[definition_name]
    else:
        definition = {
            "error": 'The policy definition named "{0}" could not be found.'.format(
                definition_name
            )
        }
    #  END

    if "error" not in definition:
        definition_id = str(definition["id"])

        prop_kwargs = {"policy_definition_id": definition_id}

        policy_kwargs = kwargs.copy()
        policy_kwargs.update(prop_kwargs)

        try:
            policy_model = await hub.exec.utils.azurerm.create_object_model(
                "resource.policy", "PolicyAssignment", **policy_kwargs
            )
        except TypeError as exc:
            result = {
                "error": "The object model could not be built. ({0})".format(str(exc))
            }
            return result

        try:
            policy = polconn.policy_assignments.create(
                scope=scope, policy_assignment_name=name, parameters=policy_model
            )
            result = policy.as_dict()
        except CloudError as exc:
            await hub.exec.utils.azurerm.log_cloud_error("resource", str(exc), **kwargs)
            result = {"error": str(exc)}
        except SerializationError as exc:
            result = {
                "error": "The object model could not be parsed. ({0})".format(str(exc))
            }
    else:
        result = {
            "error": 'The policy definition named "{0}" could not be found.'.format(
                definition_name
            )
        }

    return result


async def assignment_get(hub, ctx, name, scope, **kwargs):
    """
    .. versionadded:: 1.0.0

    Get details about a specific policy assignment.

    :param name: The name of the policy assignment to query.

    :param scope: The scope of the policy assignment.

    CLI Example:

    .. code-block:: bash

        azurerm.resource.policy.assignment_get testassign \
        /subscriptions/bc75htn-a0fhsi-349b-56gh-4fghti-f84852

    """
    polconn = await hub.exec.utils.azurerm.get_client(ctx, "policy", **kwargs)
    try:
        policy = polconn.policy_assignments.get(
            policy_assignment_name=name, scope=scope
        )
        result = policy.as_dict()
    except CloudError as exc:
        await hub.exec.utils.azurerm.log_cloud_error("resource", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


async def assignments_list_for_resource_group(
    hub, ctx, resource_group, **kwargs
):  # pylint: disable=invalid-name
    """
    .. versionadded:: 1.0.0

    List all policy assignments for a resource group.

    :param resource_group: The resource group name to list policy assignments within.

    CLI Example:

    .. code-block:: bash

        azurerm.resource.policy.assignments_list_for_resource_group testgroup

    """
    result = {}
    polconn = await hub.exec.utils.azurerm.get_client(ctx, "policy", **kwargs)
    try:
        policy_assign = await hub.exec.utils.azurerm.paged_object_to_list(
            polconn.policy_assignments.list_for_resource_group(
                resource_group_name=resource_group, filter=kwargs.get("filter")
            )
        )

        for assign in policy_assign:
            result[assign["name"]] = assign
    except CloudError as exc:
        await hub.exec.utils.azurerm.log_cloud_error("resource", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


async def assignments_list(hub, ctx, **kwargs):
    """
    .. versionadded:: 1.0.0

    List all policy assignments for a subscription.

    CLI Example:

    .. code-block:: bash

        azurerm.resource.policy.assignments_list

    """
    result = {}
    polconn = await hub.exec.utils.azurerm.get_client(ctx, "policy", **kwargs)
    try:
        policy_assign = await hub.exec.utils.azurerm.paged_object_to_list(
            polconn.policy_assignments.list()
        )

        for assign in policy_assign:
            result[assign["name"]] = assign
    except CloudError as exc:
        await hub.exec.utils.azurerm.log_cloud_error("resource", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


async def definition_create_or_update(
    hub, ctx, name, policy_rule, **kwargs
):  # pylint: disable=invalid-name
    """
    .. versionadded:: 1.0.0

    Create or update a policy definition.

    :param name: The name of the policy definition to create or update.

    :param policy_rule: A dictionary defining the
        `policy rule <https://docs.microsoft.com/en-us/azure/azure-policy/policy-definition#policy-rule>`_.

    CLI Example:

    .. code-block:: bash

        azurerm.resource.policy.definition_create_or_update testpolicy '{...rule definition..}'

    """
    if not isinstance(policy_rule, dict):
        result = {"error": "The policy rule must be a dictionary!"}
        return result

    polconn = await hub.exec.utils.azurerm.get_client(ctx, "policy", **kwargs)

    # Convert OrderedDict to dict
    prop_kwargs = {"policy_rule": loads(dumps(policy_rule))}

    policy_kwargs = kwargs.copy()
    policy_kwargs.update(prop_kwargs)

    try:
        policy_model = await hub.exec.utils.azurerm.create_object_model(
            "resource.policy", "PolicyDefinition", **policy_kwargs
        )
    except TypeError as exc:
        result = {
            "error": "The object model could not be built. ({0})".format(str(exc))
        }
        return result

    try:
        policy = polconn.policy_definitions.create_or_update(
            policy_definition_name=name, parameters=policy_model
        )
        result = policy.as_dict()
    except CloudError as exc:
        await hub.exec.utils.azurerm.log_cloud_error("resource", str(exc), **kwargs)
        result = {"error": str(exc)}
    except SerializationError as exc:
        result = {
            "error": "The object model could not be parsed. ({0})".format(str(exc))
        }

    return result


async def definition_delete(hub, ctx, name, **kwargs):
    """
    .. versionadded:: 1.0.0

    Delete a policy definition.

    :param name: The name of the policy definition to delete.

    CLI Example:

    .. code-block:: bash

        azurerm.resource.policy.definition_delete testpolicy

    """
    result = False
    polconn = await hub.exec.utils.azurerm.get_client(ctx, "policy", **kwargs)
    try:
        # pylint: disable=unused-variable
        policy = polconn.policy_definitions.delete(policy_definition_name=name)
        result = True
    except CloudError as exc:
        await hub.exec.utils.azurerm.log_cloud_error("resource", str(exc), **kwargs)

    return result


async def definition_get(hub, ctx, name, **kwargs):
    """
    .. versionadded:: 1.0.0

    Get details about a specific policy definition.

    :param name: The name of the policy definition to query.

    CLI Example:

    .. code-block:: bash

        azurerm.resource.policy.definition_get testpolicy

    """
    polconn = await hub.exec.utils.azurerm.get_client(ctx, "policy", **kwargs)
    try:
        policy_def = polconn.policy_definitions.get(policy_definition_name=name)
        result = policy_def.as_dict()
    except CloudError as exc:
        await hub.exec.utils.azurerm.log_cloud_error("resource", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


async def definitions_list(hub, ctx, hide_builtin=False, **kwargs):
    """
    .. versionadded:: 1.0.0

    List all policy definitions for a subscription.

    :param hide_builtin: Boolean which will filter out BuiltIn policy definitions from the result.

    CLI Example:

    .. code-block:: bash

        azurerm.resource.policy.definitions_list

    """
    result = {}
    polconn = await hub.exec.utils.azurerm.get_client(ctx, "policy", **kwargs)
    try:
        policy_defs = await hub.exec.utils.azurerm.paged_object_to_list(
            polconn.policy_definitions.list()
        )

        for policy in policy_defs:
            if not (hide_builtin and policy["policy_type"] == "BuiltIn"):
                result[policy["name"]] = policy
    except CloudError as exc:
        await hub.exec.utils.azurerm.log_cloud_error("resource", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result



# -*- coding: utf-8 -*-
"""
Azure Resource Manager (ARM) Network Load Balancer State Module

.. versionadded:: 1.0.0

:maintainer: <devops@eitr.tech>
:maturity: new
:depends:
    * `azure <https://pypi.python.org/pypi/azure>`_ >= 4.0.0
    * `azure-common <https://pypi.python.org/pypi/azure-common>`_ >= 1.1.23
    * `azure-mgmt <https://pypi.python.org/pypi/azure-mgmt>`_ >= 4.0.0
    * `azure-mgmt-compute <https://pypi.python.org/pypi/azure-mgmt-compute>`_ >= 4.6.2
    * `azure-mgmt-network <https://pypi.python.org/pypi/azure-mgmt-network>`_ >= 4.0.0
    * `azure-mgmt-resource <https://pypi.python.org/pypi/azure-mgmt-resource>`_ >= 2.2.0
    * `azure-mgmt-storage <https://pypi.python.org/pypi/azure-mgmt-storage>`_ >= 2.0.0
    * `azure-mgmt-web <https://pypi.python.org/pypi/azure-mgmt-web>`_ >= 0.35.0
    * `azure-storage <https://pypi.python.org/pypi/azure-storage>`_ >= 0.36.0
    * `msrestazure <https://pypi.python.org/pypi/msrestazure>`_ >= 0.6.1
:platform: linux

:configuration: This module requires Azure Resource Manager credentials to be passed as a dictionary of
    keyword arguments to the ``connection_auth`` parameter in order to work properly. Since the authentication
    parameters are sensitive, it's recommended to pass them to the states via pillar.

    Required provider parameters:

    if using username and password:
      * ``subscription_id``
      * ``username``
      * ``password``

    if using a service principal:
      * ``subscription_id``
      * ``tenant``
      * ``client_id``
      * ``secret``

    Optional provider parameters:

    **cloud_environment**: Used to point the cloud driver to different API endpoints, such as Azure GovCloud. Possible values:
      * ``AZURE_PUBLIC_CLOUD`` (default)
      * ``AZURE_CHINA_CLOUD``
      * ``AZURE_US_GOV_CLOUD``
      * ``AZURE_GERMAN_CLOUD``

    Example Pillar for Azure Resource Manager authentication:

    .. code-block:: yaml

        azurerm:
            user_pass_auth:
                subscription_id: 3287abc8-f98a-c678-3bde-326766fd3617
                username: fletch
                password: 123pass
            mysubscription:
                subscription_id: 3287abc8-f98a-c678-3bde-326766fd3617
                tenant: ABCDEFAB-1234-ABCD-1234-ABCDEFABCDEF
                client_id: ABCDEFAB-1234-ABCD-1234-ABCDEFABCDEF
                secret: XXXXXXXXXXXXXXXXXXXXXXXX
                cloud_environment: AZURE_PUBLIC_CLOUD

    Example states using Azure Resource Manager authentication:

    .. code-block:: jinja

        Ensure virtual network exists:
            azurerm.network.virtual_network.present:
                - name: my_vnet
                - resource_group: my_rg
                - address_prefixes:
                    - '10.0.0.0/8'
                    - '192.168.0.0/16'
                - dns_servers:
                    - '8.8.8.8'
                - tags:
                    how_awesome: very
                    contact_name: Elmer Fudd Gantry
                - connection_auth: {{ profile }}

        Ensure virtual network is absent:
            azurerm.network.virtual_network.absent:
                - name: other_vnet
                - resource_group: my_rg
                - connection_auth: {{ profile }}

"""
# Python libs
from __future__ import absolute_import
import logging
import re

log = logging.getLogger(__name__)

TREQ = {
    "present": {
        "require": [
            "states.azurerm.resource.group.present",
            "states.azurerm.network.public_ip_address.present",
            "states.azurerm.network.virtual_network.present",
            "states.azurerm.network.virtual_network.subnet_present",
        ]
    },
}


async def present(
    hub,
    ctx,
    name,
    resource_group,
    sku=None,
    frontend_ip_configurations=None,
    backend_address_pools=None,
    load_balancing_rules=None,
    probes=None,
    inbound_nat_rules=None,
    inbound_nat_pools=None,
    outbound_nat_rules=None,
    tags=None,
    connection_auth=None,
    **kwargs,
):
    """
    .. versionadded:: 1.0.0

    Ensure a load balancer exists.

    :param name:
        Name of the load balancer.

    :param resource_group:
        The resource group assigned to the load balancer.

    :param sku:
        The load balancer SKU, which can be 'Basic' or 'Standard'.

    :param tags:
        A dictionary of strings can be passed as tag metadata to the load balancer object.

    :param frontend_ip_configurations:
        An optional list of dictionaries representing valid FrontendIPConfiguration objects. A frontend IP
        configuration can be either private (using private IP address and subnet parameters) or public (using a
        reference to a public IP address object). Valid parameters are:

        - ``name``: The name of the resource that is unique within a resource group.
        - ``private_ip_address``: The private IP address of the IP configuration. Required if
          'private_ip_allocation_method' is 'Static'.
        - ``private_ip_allocation_method``: The Private IP allocation method. Possible values are: 'Static' and
          'Dynamic'.
        - ``subnet``: Name of an existing subnet inside of which the frontend IP will reside.
        - ``public_ip_address``: Name of an existing public IP address which will be assigned to the frontend IP object.

    :param backend_address_pools:
        An optional list of dictionaries representing valid BackendAddressPool objects. Only the 'name' parameter is
        valid for a BackendAddressPool dictionary. All other parameters are read-only references from other objects
        linking to the backend address pool. Inbound traffic is randomly load balanced across IPs in the backend IPs.

    :param probes:
        An optional list of dictionaries representing valid Probe objects. Valid parameters are:

        - ``name``: The name of the resource that is unique within a resource group.
        - ``protocol``: The protocol of the endpoint. Possible values are 'Http' or 'Tcp'. If 'Tcp' is specified, a
          received ACK is required for the probe to be successful. If 'Http' is specified, a 200 OK response from the
          specified URI is required for the probe to be successful.
        - ``port``: The port for communicating the probe. Possible values range from 1 to 65535, inclusive.
        - ``interval_in_seconds``: The interval, in seconds, for how frequently to probe the endpoint for health status.
          Typically, the interval is slightly less than half the allocated timeout period (in seconds) which allows two
          full probes before taking the instance out of rotation. The default value is 15, the minimum value is 5.
        - ``number_of_probes``: The number of probes where if no response, will result in stopping further traffic from
          being delivered to the endpoint. This values allows endpoints to be taken out of rotation faster or slower
          than the typical times used in Azure.
        - ``request_path``: The URI used for requesting health status from the VM. Path is required if a protocol is
          set to 'Http'. Otherwise, it is not allowed. There is no default value.

    :param load_balancing_rules:
        An optional list of dictionaries representing valid LoadBalancingRule objects. Valid parameters are:

        - ``name``: The name of the resource that is unique within a resource group.
        - ``load_distribution``: The load distribution policy for this rule. Possible values are 'Default', 'SourceIP',
          and 'SourceIPProtocol'.
        - ``frontend_port``: The port for the external endpoint. Port numbers for each rule must be unique within the
          Load Balancer. Acceptable values are between 0 and 65534. Note that value 0 enables 'Any Port'.
        - ``backend_port``: The port used for internal connections on the endpoint. Acceptable values are between 0 and
          65535. Note that value 0 enables 'Any Port'.
        - ``idle_timeout_in_minutes``: The timeout for the TCP idle connection. The value can be set between 4 and 30
          minutes. The default value is 4 minutes. This element is only used when the protocol is set to TCP.
        - ``enable_floating_ip``: Configures a virtual machine's endpoint for the floating IP capability required
          to configure a SQL AlwaysOn Availability Group. This setting is required when using the SQL AlwaysOn
          Availability Groups in SQL server. This setting can't be changed after you create the endpoint.
        - ``disable_outbound_snat``: Configures SNAT for the VMs in the backend pool to use the public IP address
          specified in the frontend of the load balancing rule.
        - ``frontend_ip_configuration``: Name of the frontend IP configuration object used by the load balancing rule
          object.
        - ``backend_address_pool``: Name of the backend address pool object used by the load balancing rule object.
          Inbound traffic is randomly load balanced across IPs in the backend IPs.
        - ``probe``: Name of the probe object used by the load balancing rule object.

    :param inbound_nat_rules:
        An optional list of dictionaries representing valid InboundNatRule objects. Defining inbound NAT rules on your
        load balancer is mutually exclusive with defining an inbound NAT pool. Inbound NAT pools are referenced from
        virtual machine scale sets. NICs that are associated with individual virtual machines cannot reference an
        Inbound NAT pool. They have to reference individual inbound NAT rules. Valid parameters are:

        - ``name``: The name of the resource that is unique within a resource group.
        - ``frontend_ip_configuration``: Name of the frontend IP configuration object used by the inbound NAT rule
          object.
        - ``protocol``: Possible values include 'Udp', 'Tcp', or 'All'.
        - ``frontend_port``: The port for the external endpoint. Port numbers for each rule must be unique within the
          Load Balancer. Acceptable values range from 1 to 65534.
        - ``backend_port``: The port used for the internal endpoint. Acceptable values range from 1 to 65535.
        - ``idle_timeout_in_minutes``: The timeout for the TCP idle connection. The value can be set between 4 and 30
          minutes. The default value is 4 minutes. This element is only used when the protocol is set to TCP.
        - ``enable_floating_ip``: Configures a virtual machine's endpoint for the floating IP capability required
          to configure a SQL AlwaysOn Availability Group. This setting is required when using the SQL AlwaysOn
          Availability Groups in SQL server. This setting can't be changed after you create the endpoint.

    :param inbound_nat_pools:
        An optional list of dictionaries representing valid InboundNatPool objects. They define an external port range
        for inbound NAT to a single backend port on NICs associated with a load balancer. Inbound NAT rules are created
        automatically for each NIC associated with the Load Balancer using an external port from this range. Defining an
        Inbound NAT pool on your Load Balancer is mutually exclusive with defining inbound NAT rules. Inbound NAT pools
        are referenced from virtual machine scale sets. NICs that are associated with individual virtual machines cannot
        reference an inbound NAT pool. They have to reference individual inbound NAT rules. Valid parameters are:

        - ``name``: The name of the resource that is unique within a resource group.
        - ``frontend_ip_configuration``: Name of the frontend IP configuration object used by the inbound NAT pool
          object.
        - ``protocol``: Possible values include 'Udp', 'Tcp', or 'All'.
        - ``frontend_port_range_start``: The first port number in the range of external ports that will be used to
          provide Inbound NAT to NICs associated with a load balancer. Acceptable values range between 1 and 65534.
        - ``frontend_port_range_end``: The last port number in the range of external ports that will be used to
          provide Inbound NAT to NICs associated with a load balancer. Acceptable values range between 1 and 65535.
        - ``backend_port``: The port used for internal connections to the endpoint. Acceptable values are between 1 and
          65535.

    :param outbound_nat_rules:
        An optional list of dictionaries representing valid OutboundNatRule objects. Valid parameters are:

        - ``name``: The name of the resource that is unique within a resource group.
        - ``frontend_ip_configuration``: Name of the frontend IP configuration object used by the outbound NAT rule
          object.
        - ``backend_address_pool``: Name of the backend address pool object used by the outbound NAT rule object.
          Outbound traffic is randomly load balanced across IPs in the backend IPs.
        - ``allocated_outbound_ports``: The number of outbound ports to be used for NAT.

    :param connection_auth:
        A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.

    Example usage:

    .. code-block:: yaml

        Ensure load balancer exists:
            azurerm.network.load_balancer.present:
                - name: lb1
                - resource_group: group1
                - location: eastus
                - frontend_ip_configurations:
                  - name: lb1_feip1
                    public_ip_address: pub_ip1
                - backend_address_pools:
                  - name: lb1_bepool1
                - probes:
                  - name: lb1_webprobe1
                    protocol: tcp
                    port: 80
                    interval_in_seconds: 5
                    number_of_probes: 2
                - load_balancing_rules:
                  - name: lb1_webprobe1
                    protocol: tcp
                    frontend_port: 80
                    backend_port: 80
                    idle_timeout_in_minutes: 4
                    frontend_ip_configuration: lb1_feip1
                    backend_address_pool: lb1_bepool1
                    probe: lb1_webprobe1
                - tags:
                    contact_name: Elmer Fudd Gantry
                - connection_auth: {{ profile }}

    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        if ctx["acct"]:
            connection_auth = ctx["acct"]
        else:
            ret[
                "comment"
            ] = "Connection information must be specified via acct or connection_auth dictionary!"
            return ret

    if sku:
        sku = {"name": sku.capitalize()}

    load_bal = await hub.exec.azurerm.network.load_balancer.get(
        ctx, name, resource_group, azurerm_log_level="info", **connection_auth
    )

    if "error" not in load_bal:
        # tag changes
        tag_changes = await hub.exec.utils.dictdiffer.deep_diff(
            load_bal.get("tags", {}), tags or {}
        )
        if tag_changes:
            ret["changes"]["tags"] = tag_changes

        # sku changes
        if sku:
            sku_changes = await hub.exec.utils.dictdiffer.deep_diff(
                load_bal.get("sku", {}), sku
            )
            if sku_changes:
                ret["changes"]["sku"] = sku_changes

        # frontend_ip_configurations changes
        if frontend_ip_configurations:
            comp_ret = await hub.exec.utils.azurerm.compare_list_of_dicts(
                load_bal.get("frontend_ip_configurations", []),
                frontend_ip_configurations,
                ["public_ip_address", "subnet"],
            )

            if comp_ret.get("comment"):
                ret["comment"] = '"frontend_ip_configurations" {0}'.format(
                    comp_ret["comment"]
                )
                return ret

            if comp_ret.get("changes"):
                ret["changes"]["frontend_ip_configurations"] = comp_ret["changes"]

        # backend_address_pools changes
        if backend_address_pools:
            comp_ret = await hub.exec.utils.azurerm.compare_list_of_dicts(
                load_bal.get("backend_address_pools", []), backend_address_pools
            )

            if comp_ret.get("comment"):
                ret["comment"] = '"backend_address_pools" {0}'.format(
                    comp_ret["comment"]
                )
                return ret

            if comp_ret.get("changes"):
                ret["changes"]["backend_address_pools"] = comp_ret["changes"]

        # probes changes
        if probes:
            comp_ret = await hub.exec.utils.azurerm.compare_list_of_dicts(
                load_bal.get("probes", []), probes
            )

            if comp_ret.get("comment"):
                ret["comment"] = '"probes" {0}'.format(comp_ret["comment"])
                return ret

            if comp_ret.get("changes"):
                ret["changes"]["probes"] = comp_ret["changes"]

        # load_balancing_rules changes
        if load_balancing_rules:
            comp_ret = await hub.exec.utils.azurerm.compare_list_of_dicts(
                load_bal.get("load_balancing_rules", []),
                load_balancing_rules,
                ["frontend_ip_configuration", "backend_address_pool", "probe"],
            )

            if comp_ret.get("comment"):
                ret["comment"] = '"load_balancing_rules" {0}'.format(
                    comp_ret["comment"]
                )
                return ret

            if comp_ret.get("changes"):
                ret["changes"]["load_balancing_rules"] = comp_ret["changes"]

        # inbound_nat_rules changes
        if inbound_nat_rules:
            comp_ret = await hub.exec.utils.azurerm.compare_list_of_dicts(
                load_bal.get("inbound_nat_rules", []),
                inbound_nat_rules,
                ["frontend_ip_configuration"],
            )

            if comp_ret.get("comment"):
                ret["comment"] = '"inbound_nat_rules" {0}'.format(comp_ret["comment"])
                return ret

            if comp_ret.get("changes"):
                ret["changes"]["inbound_nat_rules"] = comp_ret["changes"]

        # inbound_nat_pools changes
        if inbound_nat_pools:
            comp_ret = await hub.exec.utils.azurerm.compare_list_of_dicts(
                load_bal.get("inbound_nat_pools", []),
                inbound_nat_pools,
                ["frontend_ip_configuration"],
            )

            if comp_ret.get("comment"):
                ret["comment"] = '"inbound_nat_pools" {0}'.format(comp_ret["comment"])
                return ret

            if comp_ret.get("changes"):
                ret["changes"]["inbound_nat_pools"] = comp_ret["changes"]

        # outbound_nat_rules changes
        if outbound_nat_rules:
            comp_ret = await hub.exec.utils.azurerm.compare_list_of_dicts(
                load_bal.get("outbound_nat_rules", []),
                outbound_nat_rules,
                ["frontend_ip_configuration"],
            )

            if comp_ret.get("comment"):
                ret["comment"] = '"outbound_nat_rules" {0}'.format(comp_ret["comment"])
                return ret

            if comp_ret.get("changes"):
                ret["changes"]["outbound_nat_rules"] = comp_ret["changes"]

        if not ret["changes"]:
            ret["result"] = True
            ret["comment"] = "Load balancer {0} is already present.".format(name)
            return ret

        if ctx["test"]:
            ret["result"] = None
            ret["comment"] = "Load balancer {0} would be updated.".format(name)
            return ret

    else:
        ret["changes"] = {
            "old": {},
            "new": {
                "name": name,
                "sku": sku,
                "tags": tags,
                "frontend_ip_configurations": frontend_ip_configurations,
                "backend_address_pools": backend_address_pools,
                "load_balancing_rules": load_balancing_rules,
                "probes": probes,
                "inbound_nat_rules": inbound_nat_rules,
                "inbound_nat_pools": inbound_nat_pools,
                "outbound_nat_rules": outbound_nat_rules,
            },
        }

    if ctx["test"]:
        ret["comment"] = "Load balancer {0} would be created.".format(name)
        ret["result"] = None
        return ret

    lb_kwargs = kwargs.copy()
    lb_kwargs.update(connection_auth)

    load_bal = await hub.exec.azurerm.network.load_balancer.create_or_update(
        ctx=ctx,
        name=name,
        resource_group=resource_group,
        sku=sku,
        tags=tags,
        frontend_ip_configurations=frontend_ip_configurations,
        backend_address_pools=backend_address_pools,
        load_balancing_rules=load_balancing_rules,
        probes=probes,
        inbound_nat_rules=inbound_nat_rules,
        inbound_nat_pools=inbound_nat_pools,
        outbound_nat_rules=outbound_nat_rules,
        **lb_kwargs,
    )

    if "error" not in load_bal:
        ret["result"] = True
        ret["comment"] = "Load balancer {0} has been created.".format(name)
        return ret

    ret["comment"] = "Failed to create load balancer {0}! ({1})".format(
        name, load_bal.get("error")
    )
    if not ret["result"]:
        ret["changes"] = {}
    return ret


async def absent(hub, ctx, name, resource_group, connection_auth=None, **kwargs):
    """
    .. versionadded:: 1.0.0

    Ensure a load balancer does not exist in the resource group.

    :param name:
        Name of the load balancer.

    :param resource_group:
        The resource group assigned to the load balancer.

    :param connection_auth:
        A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.

    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        if ctx["acct"]:
            connection_auth = ctx["acct"]
        else:
            ret[
                "comment"
            ] = "Connection information must be specified via acct or connection_auth dictionary!"
            return ret

    load_bal = await hub.exec.azurerm.network.load_balancer.get(
        ctx, name, resource_group, azurerm_log_level="info", **connection_auth
    )

    if "error" in load_bal:
        ret["result"] = True
        ret["comment"] = "Load balancer {0} was not found.".format(name)
        return ret

    elif ctx["test"]:
        ret["comment"] = "Load balancer {0} would be deleted.".format(name)
        ret["result"] = None
        ret["changes"] = {
            "old": load_bal,
            "new": {},
        }
        return ret

    deleted = await hub.exec.azurerm.network.load_balancer.delete(
        ctx, name, resource_group, **connection_auth
    )

    if deleted:
        ret["result"] = True
        ret["comment"] = "Load balancer {0} has been deleted.".format(name)
        ret["changes"] = {"old": load_bal, "new": {}}
        return ret

    ret["comment"] = "Failed to delete load balancer {0}!".format(name)
    return ret
